/*
Ascii IO: Converson to and from a stream of ASCII data.
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package scotthaywood.asciiIO;

import java.util.LinkedList;
import scotthaywood.asciiIO.TypeTranslatorMap;
import java.lang.reflect.*;

/**
 * Super-class for translating objects to and from
 * a stream of ASCII data.
 * Objects that can be translated in this way are:
 * Strings, primitive wrapper types, certain user-defined
 * objects (see below), and arrays of these.
 * The user-defined objects that can be translated are
 * those that implement AsciiIO.Externalizable,
 * AsciiIO.HasTranslatableFields, or AsciiIO.Translatable.
 * @see AsciiIO.Externalizable
 * @see AsciiIO.HasTranslatableFields
 * @see AsciiIO.Translatable
 *
 * @author Scott Haywood
 * @version $Revision: 1.7 $
 */
public class AsciiIO {

	/**
	 * <code>AsciiIO.Externalizable</code> indicates that
	 * the implementing class
	 * can be be written/read using the subclasses of AsciiIO.
	 * Such a class must have a no-argument constructor.
	 * <p>
	 * <code>AsciiIO.Externalizable</code> should be implemented by
	 * classes which have public fields of simple types (primitives
	 * types, Strings, and arrays of these), which can be handled
	 * by the AsciiIO subclasses without special handling.
	 * <p>
	 * If a class has fields which are <strong>not</strong>
	 * simple (as defined
	 * above), there are two alternatives: (1) It can implement
	 * <code>AsciiIO.Translatable</code>.  (2) If it is
	 * read/written only as a field of another class, that
	 * class can implement <code>AsciiIO.HasTranslatableFields</code>
	 * and provide a translator for it.
	 *
	 * @see HasTranslatableFields
	 * @see Translatable
	 */
	public static interface Externalizable {}

	/**
	 * <code>AsciiIO.HasTranslatableFields</code>
	 * indicates that the implementing class
	 * can be be written/read using the subclasses of AsciiIO,
	 * and further, that it implements TypeTranslatorMap.
	 * Such a class must have a no-argument constructor.
	 * <p>
	 * If a class has public fields which are not simple
	 * (as defined for AsciiIO.Externalizable), and which
	 * don't implement <code>AsciiIO.Translatable</code>,
	 * it can implement <code>AsciiIO.HasTranslatableFields</code>
	 * to provide translators for those fields.
	 *
	 * @see Externalizable
	 * @see Translatable
	 */
	public static interface HasTranslatableFields extends TypeTranslatorMap {}
	
	/**
	 * <code>AsciiIO.Translatable</code>
	 * indicates that the implementing class
	 * can be be written/read using the subclasses of AsciiIO,
	 * and further, that it provides a static method by
	 * which a TypeTranslator may be obtained.  The signature
	 * of the static method must be:<br>
	 * <tt>
	 * 	public static AsciiIO.TypeTranslator getTypeTranslator();
	 * </tt>
	 * Such a class must have a no-argument constructor.
	 *
	 * @see HasTranslatableFields
	 * @see Externalizable
	 */
	public static interface Translatable {}
	
	private static void sdpl(String s) {
   		System.err.println("AsciiIO: " + s);
   	}
   	
	/**
	 * Determine if a class implements
	 * AsciiIO.HasTranslatableFields or AsciiIO.Externalizable
	 *
	 * @param	cls 	the Class object
	 *
	 * @returns true if <tt>cls</tt> implements HasTranslatableFields
	 * or Externalizable, false otherwise. 
	 *
	 * @see Externalizable
	 * @see HasTranslatableFields
	 *  
	 */
	public static boolean externalizableType(Class cls) {
		//sdpl("checking class " + cls.getName());
		boolean b =
		 (AsciiIO.HasTranslatableFields.class.isAssignableFrom(cls)
		    ||
		  AsciiIO.Externalizable.class.isAssignableFrom(cls));
		//sdpl(cls.getName() + (b?" is":" is not") +
		//     " externalizable type");
		return b;
	}

	/**
	 * Determine if an object's class implements
	 * AsciiIO.HasTranslatableFields or AsciiIO.Externalizable
	 *
	 * @param	obj 	the object
	 *
	 * @returns true if the class of <tt>obj</tt> 
	 * implements HasTranslatableFields or Externalizable, false otherwise. 
	 *
	 * @see #externalizableType(Class)
	 * @see Externalizable
	 * @see HasTranslatableFields
	 *  
	 */
	public static boolean externalizable(Object obj) {
		return obj instanceof AsciiIO.HasTranslatableFields ||
			    obj instanceof AsciiIO.Externalizable;
	}

 	//
	// A stack containing TypeTranslatorMaps
	//
	protected LinkedList typeMaps = new LinkedList();
	protected void pushTypeMap(TypeTranslatorMap m) {
		typeMaps.addFirst(m);
	}
	protected void popTypeMap() {
		typeMaps.removeFirst();
	}
	protected TypeTranslatorMap getTypeMap() {
		return typeMaps.size() > 0 ?
		       (TypeTranslatorMap) typeMaps.getFirst() :
		       null;
	}
	protected TypeTranslator getTranslator(Class type) {
		TypeTranslator t = null;

		if (Translatable.class.isAssignableFrom(type)) {
			//
			// The type provides its own translator,
			// obtained by calling a static method.
			//
			String methodName = "getTypeTranslator";
			Method m;
			try {
				m = type.getMethod(methodName, null);
			}
			catch (NoSuchMethodException e) {
				sdpl("no method " + type.getName() + "." + methodName);
				return null;
			}
			if (!Modifier.isStatic(m.getModifiers())) {
				sdpl("method " + type.getName() + "." + methodName +
					" is not static");
				return null;
			}
			try {
				Object obj = m.invoke(null, null);
				if (obj == null) {
					sdpl(type.getName() + "." + methodName + "returned null");
					return null;
				}
				t = (TypeTranslator) obj; // possible cast exception
			}
			catch (Exception e) {
				sdpl("invoke of " + type.getName() + "." + methodName +
					"failed, exception " + e);
				return null;
			}
		}
		else {
			//
			// A translator might be provided by a class
			// which implements HasTranslatableFields.
			//
			TypeTranslatorMap m = getTypeMap();
			if (m != null)
				t = m.getTypeTranslator(type);
		}
		return t;
	}
	
    protected static String squote(String s) {
    	return "'" + s + "'";		
    }
    protected static String squote(char c) {
    	return squote(String.valueOf(c));
    }
    protected static String squote(Object obj) {
    	return squote(obj.toString());
    }
}
