/*
Ascii IO: Converson to and from a stream of ASCII data.
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
package scotthaywood.asciiIO;

import java.lang.reflect.*;

/**
 * A <code>BasicTypeTranslator</code> is a TypeTranslator
 * that handles one type,
 * and is constructed from a <code>Surrogate</code> object.
 *
 *
 * @author Scott Haywood
 * @version $Revision: 1.7 $
 */
public class BasicTypeTranslator implements TypeTranslator {
	
	private Class ttype;
	private Class stype;
	
	/**
	 * Constructs a new <code>BasicTypeTranslator</code>
	 * using a given <code>Surrogate</code>.
	 *
	 * @param	s	The <code>Surrogate</code>.
	 */
	public BasicTypeTranslator(Surrogate s) {
		ttype = s.getTargetType();
		stype = s.getClass();
	}
	
	/**
	 * Return the target types which are supported.
	 *
	 * @returns an array of one element, containing
	 * the <code>Class</code> object of the <code>Surrogate</code>.
	 */
	public Class[] getTargetTypes() {
		return new Class[] {ttype};
	}

	/**
	 * Return the surrogate type for a given target type.
	 *
	 * @param	targetType	the class object of the target.
	 * @returns the class object of the surrogate, or null if
	 * the target is not supported.
	 * 
	 */
	public Class getSurrogateType(Class targetType) {
		if (targetType != ttype) {
			throw new Error("got " + targetType.getName() +
			                ", expected " + ttype.getName());
		}
		return stype;
	}
	
	
	/**
	 * Return a surrogate object, given a target.
	 *
	 * @param	target	the target object.
	 * @returns the <code>Surrogate</code> object, or null if
	 * the target is not supported.
	 * 
	 */
	public Object createSurrogate(Object target) {
		Constructor con = null;
		//dpl("creating surrogate for object of type " +
		//	target.getClass().getName());
		//dpl("target obj=" + target);
		try {
			con = stype.getConstructor(
				new Class[] {ttype});
			return con.newInstance(new Object[] {target});
		}
		catch (Exception e) {
			dpl("tried constructor " + stype + "(" + ttype + ")\n" +
			    "constructor = " + con + "\n" +
			    "passing it " + target.getClass() + "\n" +
			    "caught exception: " + e + e.getMessage());
			return null;
		}
	}

	/**
	 * Return a target object, given a surrogate.
	 *
	 * @param	surrogate	the <code>Surrogate</code> object.
	 * @param	cls		the desired type of target
	 * @returns the target object, or null if the target type
	 * is not supported.
	 * 
	 */
	public Object createTarget(Class cls, Object surrogate) {
		if (cls != ttype) return null;
		Surrogate s = (Surrogate) surrogate;
		return s.createTarget();
	}
	
    private void dpl(String s) {
		System.err.println(getClass().getName() + ": " + s);
	}
	
}
