/*
Ascii IO: Converson to and from a stream of ASCII data.
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
package scotthaywood.asciiIO;

/**
 * Converts ascii strings that have embedded control-sequences
 * and Unicode escapes.
 * <p>
 * Example:<br>
 * <code>ConvertUnicode.fromAscii("hello\\n\\uFEDC", "\\u");</code><br>
 * produces a string "hello\n" + (char)0xFEDC;<br>
 * 
 *
 * @author Scott Haywood
 * @version $Revision: 1.6 $
*/
public class ConvertUnicode {

//
// Convert the control-sequences in the given ascii string.
// A control-sequence is a back-slash followed by one of
// b, f, n, r, or t.
//
private static String controlCharsFromAscii(String s) {
	StringBuffer sb = new StringBuffer();
	int start = 0;
	int pos;
	while ((pos=s.indexOf('\\', start)) != -1) {
	    sb.append(s.substring(start, pos));
	    if (pos == s.length() -1) {
	    	sb.append('\\');
	    	start = s.length();
	    	break;
	    }
	    else {
	    	boolean isOctal = false;
	    	int v;
	    	char newchar = (char)0;
	    	start = pos+1;
	    	char char2 = s.charAt(start);
	    	while ((v = Character.digit(char2, 8)) != -1) {
	    		isOctal = true;
	    		newchar = (char)((char)(newchar << 3) + v);
	    		start++;
	    		if (start == pos+4) break;
	    		char2 = s.charAt(start);
	    	}
    		if (!isOctal) {
	    		start = pos+2;
	    		switch (char2) {
	    			case 'b' : sb.append('\b'); break;
	    			case 'f' : sb.append('\f'); break;
	    			case 'n' : sb.append('\n'); break;
	    			case 'r' : sb.append('\r'); break;
	    			case 't' : sb.append('\t'); break;
	    			default  : sb.append(char2); break;
	    		}
	    	}
	    }
	}
	if (start < s.length())
		sb.append(s.substring(start));
	return sb.toString();
}
/**
 * Convert the Unicode sequences in the given ASCII string.
 * Each sequence consists of the given escape string,
 * followed by 1 to 4 hex digits.
 * After converting Unicode sequences, convert some
 * control-character sequences, for example "\b" is converted
 * to a backspace.
 * @param	s	the string to be converted.
 * @param	escape	the escape sequence that prefixes Unicode sequences.
 * @returns	the converted string.
*/
public static String fromAscii(String s, String escape) {

	StringBuffer rv = new StringBuffer();
	int start = 0;
	int pos;
	
	while ((pos=s.indexOf(escape, start)) != -1) {
    	rv.append( s.substring(start,pos) );
    	char newchar = 0;
    	int i;
    	for (i=pos + escape.length();
    	     (i < pos + escape.length()+4) && i<s.length();
    	     i++)
    	{
    		int v;
    		if ((v = Character.digit(s.charAt(i), 16)) != -1) {
    			newchar = (char)((char)(newchar << 4) + v);
    		}
    		else break;
    	}
    	if (i == pos + escape.length() ) {
    		// no valid sequence to be converted,
    		// so insert the escape as a literal.
    		// Note that it might be altered outside this loop.
    		rv.append( escape );
    	}
    	else {
    		//System.err.println("fromAscii: char value:" + (int)newchar);
    		rv.append( newchar );
    	}
    	start = i;
    }
    rv.append( s.substring(start) );
    return controlCharsFromAscii(rv.toString());
}

} //end class
