/*
Ascii IO: Converson to and from a stream of ASCII data.
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
package scotthaywood.asciiIO;

import java.lang.reflect.*;
import java.io.*;
import java.util.*;
import scotthaywood.asciiIO.AsciiIO;
import scotthaywood.asciiIO.ConvertUnicode;
import scotthaywood.asciiIO.Surrogate;
import scotthaywood.asciiIO.TypeTranslatorMap;
import scotthaywood.asciiIO.TypeTranslator;


/**
 A <code>DataReader</code> can be used to read data items
 from a java.io.Reader.  The data items can
 be of any primitive or String types,
 or arrays of these of any number of dimensions.

 Example:
 Assume the text file "file_x" looks like this --

    5
    hello
    { 1, 2, 3 }
    true

 Then the code to read in the data could be --

 	DataReader dp = new DataReader(new FileReader("file_x"));
 	Integer x = (Integer) dp.read(int.class); // or Integer.class
 	String s = (String) dp.read(String.class);
 	int[] intarray = (int[]) dp.read(int[].class);
 	Boolean b = (Boolean) dp.read(boolean.class);

 Note that read() always returns an Object, so read(int.class)
 will return an Integer rather than a int.  On the other hand,
 an array is already an object, so read(int[].class) will
 return an array of int.

 In the data file, arrays can be grouped using curly brackets,
 but brackets are not always needed. Array items can be separated
 by whitespace or commas.  Consecutive commas are ignored, so that
 	0, , 2
 when read as an array, is the same as
 	0, 2
 or
 	0
 	2
 	
 The "file_x" example above could have been read
 like this --

    DataReader dp = new DataReader(new FileReader("file_x"));
 	String five_str = (String) dp.read(String.class);
	String hello_str = (String) dp.read(String.class);
  	byte[] bytearray = dp.read(byte[].class);
 	String b = (String) dp.read(String.class);

 Errors in reading or converting are thrown as DataReaderError,
 defined herein.
   	
 *
 * @author Scott Haywood
 * @version $Revision: 1.10 $
*/
public class DataReader extends AsciiIO {

	StreamTokenizer st;
	final char quote = '"';
	final char openBracket = '{';
	final char closeBracket = '}';
	final char comma = ',';
	final char equal = '=';
	
	HashMap primTypes = new HashMap();
	
	private void dpl(String s) {
		String base=getClass().getName();
		base=base.substring(base.lastIndexOf('.')+1);
   		System.err.println(base + ": " + s);
   	}
   	
	public DataReader(Reader r) {
	
		st = new StreamTokenizer(r);
		st.resetSyntax();
		st.whitespaceChars(' ',' ');
		st.whitespaceChars('\t','\r');
		st.wordChars('!','~');
		// these must be after wordChars() ...
		st.commentChar('#');
		st.quoteChar(quote);
		st.lowerCaseMode(false);
		st.slashSlashComments(false);
		st.slashStarComments(false);
		st.eolIsSignificant(false); // if true, must also be in whitespaceChars
		st.ordinaryChar(openBracket);
		st.ordinaryChar(closeBracket);
		st.ordinaryChar(comma);
		st.ordinaryChar(equal);
		
		primTypes.put(boolean.class, Boolean.class);
		primTypes.put(byte.class, Byte.class);
		primTypes.put(char.class, Character.class);
		primTypes.put(double.class, Double.class);
		primTypes.put(int.class, Integer.class);
		primTypes.put(float.class, Float.class);
		primTypes.put(long.class, Long.class);
		primTypes.put(short.class, Short.class);
	}
	
	
	// Read an object of the given type.
	// Return null in case of error.
	// This method is private because it will leave
	// the input stream in an unknown state unless
	// the expected type is represented by a token.
	private Object readIfAny(Class expected) {
		Object obj = null;
		try { obj = read(expected); }
		catch (DataReaderError d) {}
		return obj;
	}
	
	/**
	 * Reads a string from the imput stream.
	 * @return the string, if a bareword or quoted string
	 * is successfully read from the stream, <tt>null</tt> otherwise.
	 */
	public String readStringIfAny() {
		return (String) readIfAny(String.class);
	}
	

	/** Read and return a value of a given type.
	 *
	 *  @param	expected 	class object of the expected type
	 *
	 *  @return an object of the expected type, decoded from
	 *  the imput stream.  If the type is a primitive type,
	 *  returns an object of the corresponding
	 *  wrapper type (e.g. an Integer is returned if "int" was requested).
	 *
	 *  @throws <tt>DataReaderError</tt> if not able to read
	 *  the expected type.
	 *
	 */
	public Object read(Class expected) {
	
		Object rv = null;
		TypeTranslator trans = null;
		if (lookForChar(closeBracket)) {
			st.pushBack();
			// return value remains null
		}
		else if (expected.isArray()) {
        	rv = readArray(expected);
		}
		else if (expected.isPrimitive()) {
			rv = readPrimitive(expected);
		}
		else if (expected.equals(String.class)) {
			rv = readString();	
		}
		else if ((trans = getTranslator(expected)) != null) {
			Class surrType = trans.getSurrogateType(expected);
			Object surr = read(surrType);
			rv = trans.createTarget(expected, surr);
		}
		else if (externalizableType(expected)) {
			boolean is_null = false;
			int tok = nextToken();
			switch (tok) {
			case StreamTokenizer.TT_WORD:
				// unquoted literal null
				if (st.sval.equals("null")) {
					is_null = true;
				}
				// fall thru
			default:
				st.pushBack();
				break;
			}
			rv = is_null ? null : readPublicFieldsOfType(expected);
		}
		else {
			throw new DataReaderError("don't know how to parse type "+
			                          squote(expected.getName()));
		}
		return rv;
	}

	
	// read an array of the given type
	//
	private Object readArray(Class expected) {
		if (!expected.isArray()) {
			throw new DataReaderError("not an array type: "+
			                          squote(expected.getName()));
		}
		// the token "null", without quotes, represents a null value
		int tok = nextToken();
		boolean isNull = false;
		switch (tok) {
			case StreamTokenizer.TT_WORD:
				if (st.sval.equals("null"))
					isNull = true;
				else
					st.pushBack();
				break;
			default:
				st.pushBack();
				break;
		}
		Object rv = null;
		if (isNull) {
			// do nothing
		}
		else if (expected == char[].class) {
			rv = readCharArray();
		}
		else {
			boolean bracket = lookForChar(openBracket);
			rv = readList(expected.getComponentType());
			if (bracket && ! lookForChar(closeBracket) ) {
				throw new DataReaderError("expected " + closeBracket);
			}
		}
		return rv;	
	}
		
	//
	// Read a word or quoted string,
	// convert it to the wrapper type that
	// corresponds to the type given as a parameter.
	//
	// If the requested type is char, only the
	// first character of the string is used.
	//
	private Object readPrimitive(Class expected) {
		if (!expected.isPrimitive()) {
			throw new DataReaderError("not a primitive type: "+
			                          squote(expected.getName()));
		}
		Class wrapper = (Class) primTypes.get(expected);
		int tok = nextToken();
		String sss = null;
		boolean badType = false;
		switch (tok) {
			case quote:
			case StreamTokenizer.TT_WORD:
				sss = st.sval;
				break;
			case StreamTokenizer.TT_NUMBER:
				sss = String.valueOf(st.nval);
				break;
			default:
				badType = true;
				break;
		}
		if (badType) throw new DataReaderError("expected type:" +
				                               expected.getName() +
				                                ", read " + tokenName() );
		
		Object rv = null;
		if (expected == char.class || expected == Character.class) {
			if (sss.length() == 0)
				throw new DataReaderError("can't convert null string " +
				                          " to Character");
			rv = new Character( convertUnicode(sss).charAt(0) );
		}
		else {
    		Constructor con = null;
    		try { con = wrapper.getConstructor(new Class[] { String.class }); }
    		catch (Exception e) {
    			/* should never be here */
    			throw new DataReaderError(e, e.toString());
    		}
    		try { rv = con.newInstance( new Object[]{sss} ); }
    		catch (Exception e) {
    			throw new DataReaderError(e, "could not convert string "+
    			                          squote(sss) + " to type " + wrapper);
    		}
		}	
		return rv;
	}
	
	
	
	/**
	 * Read a string, which may be in the form of
	 * a quoted string or a word token (a bare word).
	 * 
	 * @returns the string, or null if the unquoted word null is read.
	 * @throws <tt>DataReaderError</tt> if a string cannot be read.
	 * 
	 */
	public String readString() {
		int tok = nextToken();
		switch (tok) {
			case quote:
				return convertUnicode(st.sval);
			case StreamTokenizer.TT_NUMBER:
				return Double.toString(st.nval);	
			case StreamTokenizer.TT_WORD:
				String sss = st.sval;
				return (sss.equals("null")) ? null :
					    convertUnicode(sss);
			default:
				throw new DataReaderError("expected a string, read " +
				                          tokenName() );
		}	
	}
	
	// read an char[] array. In the input stream it
	// should look like a string.
	private char[] readCharArray() {
		String s = readString();
		return s.toCharArray();
	}
	
	//
	// Read a comma- or space-separated list of
	// objects of a given type,
	// and return them in an array.
	//
	private Object readList(Class expected) {
		ArrayList list = new ArrayList();
		Object elem = null;
	   loop: while (true) {
			int tok = nextToken();
			switch (tok) {
				case comma: // treated as whitespace
					break;
				case closeBracket:
					st.pushBack();
					break loop;
				default:
					st.pushBack();
					elem = read(expected);
					if (elem == null) break loop;
					list.add(elem);
					break;
			}
		}
		// convert ArrayList to an array of the expected type.
		// Note that a primitive type such as "int" can be set
		// via Array.set() using a value of the corresponding
		// wrapper class (such as Integer).
		Object array = Array.newInstance(expected, list.size());
		for (int i=0; i<list.size(); i++) {
			try { Array.set(array, i, list.get(i)); }
			catch (Exception e) {
				throw new DataReaderError(e + "can't set array of type " +
				                          squote(expected.getName()));
			}
				
		}
		return array;
	}
	
	// Create an object of the given type and read its fields.
	//
	private Object readPublicFieldsOfType(Class cls) {
		Object obj = null;
		try { obj = cls.newInstance(); }
		catch (Exception e) {
			throw new DataReaderError("! can't instantiate " + cls.getName()
				+ ":\n"
				+ "exception type = " + e.getClass().getName() + "\n"
				+ " !\n"
				+ e.toString()
				+ " !\n"
				);
		}
		readPublicFields(obj);
		return obj;
	}
	
	/**
	 * Interpret the data stream as values to
	 * be stored in the public fields of the given object.
	 * If possible use TypeTranslators to read fields with types
	 * that are not primitives or Strings.
	 *
	 * @param	obj		the object whose fields are to be set.
	 * 
	 * @throws <tt>DataReaderError</tt> if not able to read
	 * the expected type.
	 *
	 */
	public void readPublicFields(Object obj) {
		boolean is_map = obj instanceof HasTranslatableFields;
		if (is_map) pushTypeMap((TypeTranslatorMap) obj);
		try { readPublicFields0(obj); }
		finally { if (is_map) popTypeMap(); }
	}

	
	// Interpret the data stream as values to
	// be stored in the public fields of the given object.
	//
	private void readPublicFields0(Object obj) {
		Class cls;
		cls = obj.getClass();
		boolean bracket = false;
		if (lookForChar(openBracket))
			bracket = true;
		loop: while (true) {
			int tok = nextToken();
			switch (tok) {
			case StreamTokenizer.TT_WORD:
				String fieldName = st.sval;
				//dpl("<<fieldName="+fieldName+">>");
				if (!lookForChar(equal)) {
					throw new DataReaderError(
					          "error reading fields for object of type " +
			                  squote(cls.getName()) + ", expected " +
			                  squote(equal));
			    }
			    Field f = null;
			    try { f = cls.getField(fieldName); }
			    catch (Exception e) {
			    	throw new DataReaderError(e,
			    		"error reading type '" + cls.getName() + "':\n" +
			    		"data stream contains unknown field '" +
			    		 fieldName + "'");
			    }
			    if (!Modifier.isPublic(f.getModifiers())) {
			    	throw new DataReaderError(
			    		"field " + squote(f.getName()) +
			    		" of class " + squote(cls.getName()) +
			    		" is not public");
			    }
			    Class fieldType = f.getType();
			    Object value = read(fieldType);
			    /*
			    dpl("<<got value " + squote(value) +
			    	 " for field " + f.getName() +
			    	 " of type " + fieldType.getName() + ">>");
			    */
			    try { f.set(obj, value); }
			    catch (Exception e) {
			    	throw new DataReaderError(e,
			    		"error reading type '" + cls.getName() + "':\n" +
			    		"cannot assign value (" +
			    		(value==null ? "null" : value.toString()) +
			    		") to field '" + fieldName + "' " +
			    		"of type '" + fieldType + "'");
			    }
			    break;
			default:
				st.pushBack();
				break loop;
			}	
		}
		if (bracket && ! lookForChar(closeBracket) ) {
			throw new DataReaderError("expected " + closeBracket);
		}
		
	}
	
	private boolean lookForChar(char c) {
		int tok = nextToken();
		if (tok == c) { return true; }
		else { st.pushBack(); return false; }
	}
	
	private int nextToken() {
		int rv = 0;
		try { rv = st.nextToken(); }
		catch (Exception e) { throw new DataReaderError(e, "read failed"); }
		return rv;
	}
	
	// Return a representation of the last read token.
	private String tokenName() {
		switch (st.ttype) {
		case StreamTokenizer.TT_EOF :    return "EOF";
		case StreamTokenizer.TT_EOL :    return "EOL";
		case StreamTokenizer.TT_NUMBER : return "NUMBER:" + st.nval;
		case StreamTokenizer.TT_WORD :   return "STRING:" + st.sval;
		default: return "CHAR:" + st.ttype +
		                (isPrintableAscii(st.ttype) ?
		                 ("("+ (char)st.ttype + ")") : "");
		}
	}
	
	static boolean isPrintableAscii(int c) {
		return (c >= ' ') && (c <= '~');
	}
	
	private String convertUnicode(String s) {
		String rv = ConvertUnicode.fromAscii(s, "\\u");
		//dpl("convert '" + s + "' to '" + rv + "'");
		return rv;
	}
	
	/**
	 * <code>DataReaderError</code> object is thrown when an
	 * i/o or parsing error occurs.
	 *
	 */
	public class DataReaderError extends Error {
		Throwable throwable;

		/**
		 * Get the underlying error (for example, i/o error).
		 * @returns the error.
		 *
		 */
		public Throwable getThrowable() { return throwable; }

		private DataReaderError() { super(); }
		private DataReaderError(Throwable t, String s) {
			super("line " + st.lineno() + ":" + s);
			throwable = t;
		}
		private DataReaderError(String s) {
			super("line " + st.lineno() + ":" + s);
		}
	}
}
