/*
Ascii IO: Converson to and from a stream of ASCII data.
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
package scotthaywood.asciiIO;


import java.io.*;
import java.util.HashSet;
import java.lang.reflect.*;
import scotthaywood.asciiIO.TypeTranslatorMap;
import scotthaywood.asciiIO.TypeTranslator;
import scotthaywood.asciiIO.Surrogate;

/**
 *   This class writes data in a form that can be
 *   read by the DataReader class
 *  
 *   Usage:
 *   <tt>
 *  	DataWriter dw = new DataWriter(new FileWriter(filename));
 *    dw.write(obj);
 *	</tt>
 *   where obj is an object consisting of primitive types (char, int, etc.)
 *   and their wrapper types (Character, Integer, etc.), Strings, and
 *   arrays of these types.
 *   Certain user-defined types are also supported; see AsciiIO.
 *   <p>
 *   Examples:
 *   <tt>
 *    write(new int[] {1,2,3});  // writes seven characters: {1,2,3}
 *    write(new char[] {'a', 'b'});  // writes "ab" (including the quotes)
 *	</tt>
 *  
 *  @see AsciiIO
 *  @see DataReader
 *
 * @author Scott Haywood
 * @version $Revision: 1.9 $
 */
public class DataWriter extends AsciiIO {

	private static void vprintln(String s) {
		System.err.println(s);
	}
	
	private Writer writer;
	private static HashSet primWrappers = new HashSet();
	private static String[] charReps = new String[(int)' '];
	static {
		primWrappers.add(Boolean.class);
		primWrappers.add(Byte.class);
		primWrappers.add(Character.class);
		primWrappers.add(Double.class);
		primWrappers.add(Float.class);
		primWrappers.add(Integer.class);
		primWrappers.add(Long.class);
		primWrappers.add(Short.class);
		charReps['\n'] = "\\n";
		charReps['\b'] = "\\b";
		charReps['\t'] = "\\t";
		charReps['\f'] = "\\f";
		charReps['\r'] = "\\r";
	}
	
	public DataWriter(Writer w) {
		writer = w;
	}
	
	/*
	*  All characters sent to the java.io.Writer
	*  go through these put() methods.
	*/
	private int nwritten = 0;
	private char lastwritten;
	private final char newline = '\n';
	
	private void put(String s) {
		try { writer.write(s); }
		catch (IOException e) {
			throw new DataWriterError(e, "write failed");
		}
		int len = s.length();
		nwritten += len;
		if (len > 0)
			lastwritten = s.charAt(len - 1);
	}
	
	private void put(char c) {
		try { writer.write(c); }
		catch (IOException e) {
			throw new DataWriterError(e, "write failed");
		}
		nwritten++;
		lastwritten = c;
	}
	
	private boolean isStartOfLine() {
		return nwritten == 0 || lastwritten == newline;
	}
	
	private void ensureNewLine() {
		if (!isStartOfLine()) put(newline);
	}
	
	/*
	*  Public methods
	*/
	
	/**
	 * Flushes the output stream. 
	 */
	public void flush() {
		try { writer.flush(); }
		catch (IOException e) {
			throw new DataWriterError(e, "flush failed");
		}
	}
	
	
	/**
	 * Writes the public fields of an object to the output stream.
	 * See AsciiIO for a description of the types of fields that
	 * this method can handle.
	 * @see AsciiIO
	 */
	public void write(Object obj) {
		write(obj, -1);
	}	
	
	private void write(Object obj, int level) {
		level++;
		if (obj == null) {
			put("null");
			return;
		}
		Class type = obj.getClass();
		TypeTranslator trans = null;
		
		if (type == String.class) {
			writeString((String) obj);
		}
		else if (primWrappers.contains(type)) {
			if (type.equals(Character.class)) {
				writeString(obj.toString());
			}
			else {
				put(obj.toString());
			}
		}
		else if (type.equals(char[].class)) {
			writeString( new String((char[])obj) );
		}
		else if (type.isArray()) {
			int len = 0;
			try { len = Array.getLength(obj); }
			catch (Exception e) {
				throw new DataWriterError("can't get array length");
			}
			put('{');
			for (int i=0; i<len; i++) {
				Object elem = null;
				try { elem = Array.get(obj, i); }
				catch (Exception e) {
					throw new DataWriterError("can't get array element");
				}
				write(elem, level);
				if (i < len-1)
					put(',');
			}
			put('}');
		}
		else if ((trans = getTranslator(type)) != null) {
			Object surr = trans.createSurrogate(obj);
			write(surr, level);
		}
        else if (externalizable(obj)) {
        	writePublicFields(obj, level);
		}
		else {
			throw new DataWriterError("don't know how to write type "+
			                          squote(type.getName()));
		}
		
	}
	
	/**
	 * Converts a string to a canonical form: adding quotes
	 * as needed so that the string can be properly read by
	 * the DataReader class.
	 * @param	s	the string to be converted.
	 * @returns a canonical version of the string.
	 * @see #writeString
	 */
	public static String canonicalString(String s) {
		if (s == null) return "null";
		StringBuffer sb = new StringBuffer();
		boolean addQuotes = false;
		for (int i=0; i<s.length(); i++) {
			char c = s.charAt(i);
			if (c < ' ') {
				String rep = charReps[c];
				if (rep != null)
					sb.append(rep);
				else {
					sb.append('\\');
					// or maybe: sb.append(Integer.toString((int)c, 0));
					sb.append(Character.forDigit((c>>>6) & 007, 8));
					sb.append(Character.forDigit((c>>>3) & 007, 8));
					sb.append(Character.forDigit( c      & 007, 8));
				}
			}
			else if (c > 0xff) { // unicode escape: \\uXXXX
				sb.append("\\\\u");
				// or maybe: sb.append(Integer.toString((int)c, 16));
				sb.append(Character.forDigit((c>>>12) & 0xf, 16));
				sb.append(Character.forDigit((c>>> 8) & 0xf, 16));
				sb.append(Character.forDigit((c>>> 4) & 0xf, 16));
				sb.append(Character.forDigit( c       & 0xf, 16));
			}
			else {
				if (c=='"' || c=='\\')
					sb.append('\\');
				else if (c == ' ')
					addQuotes = true;
				sb.append(c);
			}
		}
		if (addQuotes || sb.length()==0) {
			sb.insert(0,'"');
			sb.append('"');
		}
		return sb.toString();
	}
	
	/**
	 * Writes a string to the output stream.
	 */
	public void writeString(String s) {
		put(canonicalString(s));
	}
	
	/**
	 * Writes the public fields of an object to the output stream.
	 * @see #write
	 */
	public void writePublicFields(Object obj) {
		writePublicFields(obj, -1);
	}
	
	private void writePublicFields(Object obj, int level) {
		boolean is_map = obj instanceof TypeTranslatorMap;
		if (is_map) pushTypeMap((TypeTranslatorMap) obj);
		try {
			if (level>0) put('{');
			writePublicFields0(obj, level);
			if (level>0) put('}');
		}
		finally { if (is_map) popTypeMap(); }
	
	}
	
	private void writePublicFields0(Object obj, int level) {
		Class cls;
		Field[] fields;
		cls = obj.getClass();
		try { fields = cls.getFields(); }
		catch (Exception e) {
			throw new DataWriterError("could not get fields for object of type " +
			                          squote(cls.getName()));
		}
		ensureNewLine();
		writeComment("class " + cls.getName() + ", " +
		             fields.length + " public field" +
		             (fields.length == 1 ? "" : "s"));
		for (int i=0; i<fields.length; i++) {
			Field f = fields[i];
			int mods = f.getModifiers();
			//vprintln("writePublicFields: field " + f.getName());
			
			// getFields should return only public fields, but check anyway
			if (! Modifier.isPublic(mods) ) continue;
			String fieldName = f.getName();
			//vprintln("writePublicFields: got field " + fieldName);
			Class type = f.getType();
			Object value = "unknown";
			try { value = f.get(obj); }
			catch (Exception e) {
				throw new DataWriterError(e,
					"could not get value of field " +
				    squote(fieldName) +
				    " from object of type " +
			        squote(cls.getName()));
			}
			//vprintln(fieldName + "=" + value);
			ensureNewLine();
			put(fieldName + "=");
			write(value, level);
		}
		ensureNewLine();
	}
	
	
	/**
	 * Writes a comment to the output stream
	 */
	public void writeComment(String s) {
		if (! isStartOfLine()) put(' ');
		int pos;
		while ((pos = s.indexOf(newline, 0)) != -1) {
			 put("# " + s.substring(0, pos+1));
			 s = s.substring(pos+1);
		}
		if (s.length() != 0)
			put("# " + s + newline);
	}
	
	/**
	 * A <code>DataWriterError</code> is the thrown by
	 * <code>DataWriter</code> when it fails.
	 */
	public class DataWriterError extends Error {
		Exception exception = null;
		public DataWriterError() { super(); }
		public DataWriterError(String s) {
			super(/* "line " + lineno() + ":" + */ s);
		}
		public DataWriterError(Exception e, String s) {
			super(/* "line " + lineno() + ":" + */ s);
			exception = e;
		}
		public Exception getException() {
			return exception;
		}
	}
	
	
}
