/*
Image Creator: Creates configurable images
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
package scotthaywood.imageCreator;

import java.lang.reflect.*;
import java.util.*;
import java.io.*;
import scotthaywood.imageCreator.Module;
import scotthaywood.imageCreator.ModuleConfig;
import scotthaywood.asciiIO.*;

/**
	This class manages the configuration file
	for an application.
	
	The format for the configuration file is --
		[ header-data ]
		[
		$module mod-name [ mod-class ]
		[ module-data ]
		] ...
	where $module is a literal; mod-name is a string as
	written/read by the AsciiIO classes; mod-class is the name
	of a subclass of Module (equals mod-name if not given);
	header-data and module-data are text written and read
	by the AsciiIO classes.
	
 *
 * @author Scott Haywood
 * @version $Revision: 1.12 $
*/
public class Conf {

	/**
	 * Constructs a configuration object given the name of
	 * a file which contains the data.
	 * @param	fileName	the file name
	 */
	public Conf(String fileName) throws IOException {
		this(new FileReader(fileName));
		filename = fileName;
	}
	/**
	 * Constructs a configuration object given a
	 * <code>java.io.Reader</code> which contains the data.
	 * @param	reader	the Reader
	 */
	public Conf(Reader reader) throws IOException {
		myName = this.getClass().getName();
		LineNumberReader lnr = new LineNumberReader(reader);
		StringBuffer sb = new StringBuffer();
		String modname = null;
		String clsname = null;
		String line;
		
		while ((line = lnr.readLine()) != null) {
			if (line.startsWith(mkeyword)) {
				if (modname == null) {
					String headerText = sb.toString();
					try { readModuleConfig(headerText, header); }
					catch (Throwable t) {
						dpl("error reading header, text='" + headerText + "',\n" +
						"got error " + t.toString() +
						" error message :" + t.getMessage());
					}
				}
				else {
					addMod(modname, clsname, sb.toString());
				}
				sb.delete(0,Integer.MAX_VALUE);
				DataReader dr = new DataReader(
					new StringReader(line.substring(mkeyword.length())));
				modname = dr.readString();
				if (modname == null || modname.length() == 0) {
					throw new IOException(myName +
					                      "null module name, line " +
					                      lnr.getLineNumber());
				}
				clsname = dr.readStringIfAny();
				if (clsname == null || clsname.length() == 0) {
					clsname = modname;
				}
			} else {
				// note that readline() above discards
				// any linefeed or CR-LF.
				sb.append(line + linesep);
			}
					
		}
		if (modname != null)
			addMod(modname, clsname, sb.toString());
	}

    /**
    	The Header object contains all the
    	configuration data that is not specific
    	to any particular module
    */
    public class Header extends ModuleConfig {
    	/**
    	 * The name of the current module
    	 */
    	public String selected_module;
    	
    	/**
    	 * List of names of packages in which modules
    	 * may be found.
    	 * A null string denotes the empty (default) package.
    	 *
    	 */
    	public String[] module_packages = { "" };
    	
    	/**
    	 * Convert this object to a String, for debugging only
    	 */
    	public String toString() {
    		String p1="selected_module="+selected_module;
    		String p2="module_packages=";
    		if (module_packages==null)
    			p2 += "null";
    		else if (module_packages.length==0)
    			p2 += "empty";
    		else {
    			p2 += "'" +module_packages[0] + "'";
    			for (int i=1; i<module_packages.length; i++) {
    				p2 += ",'" + module_packages[i] + "'";
    			}
    		}
    		return "Conf.Header["+p1+";"+p2+"]";
    			
    	}
    	
    }
	
	/**
	 * A module entry in the configuration
	 */
	 private class ModDescr {

	 	//
		String name;

		// the class name, relative to the package list.
		// This is the name given to loadModuleClass()
		String loadName;

		// The full class name (not relative to package list).
		// May be unknown.
		//String fullClassName = "";

		// the configuration text
		String text;
		
		// number attempted to add with this name
		int count = 1;
		
		public ModDescr(String n, String l, String t) {
			name=n; loadName=l; text=t;
		}
	}
	
	/**
	 * Add an entry to the list of modules,
	 * ensuring that it has a unique name.
	 * @param name 	the desired name
	 * @param clsname 	the class name; used for loading
	 * @param text 	the configuration text
	 * @returns the actual name added, which is based on
	 * the desired name.
	 */
	private String addMod(String name,
	                      String clsname,
						  String text) {
		//dpl("adding " + name);
		ModDescr md = new ModDescr(name, clsname, text);
		boolean collide = mods.get(name) != null;
		if (collide) {
			ModDescr first = (ModDescr) mods.get(clsname);
			if (first == null)
				md.name = clsname;
			else {
				while (collide) {
					first.count++;
					md.name = name + " <" + first.count + ">";
					collide = (mods.get(md.name) != null);
				}
			}
		}
		modNames.add(md.name);
		mods.put(md.name, md);
		return md.name;
	}
	
	private String getModText(String name) {
		ModDescr md = (ModDescr) mods.get(name);
		return md == null ? null : md.text;
	}
	
    	
	public String[] getModNames() {
		String[] sa = new String[modNames.size()];
		Object[] oa = modNames.toArray();
		System.arraycopy(oa,0,sa,0,oa.length);
		//for (int i=0; i<sa.length; i++)
		//	dpl("sa[" + i + "]=" + sa[i]);
		return sa;
	}
		
	/**
	 * Tells whether a module name is in the configuration
	 * @param	modName	the name
	 * @returns	true if the name exists, false if not.
	 */
	public boolean hasModule(String modName) {
		return mods.get(modName) != null;
	}
	/**
	 * Reads a module's configuration from this object.
	 * @param	modName	the name
	 * @param	m	a ModuleConfig object to be overwritten
	 * with the module's configuration
	 * @returns	true on success, false on failure.
	 */
	public boolean read(String modName, ModuleConfig mc) {
		ModDescr md = (ModDescr) mods.get(modName);
		if (md == null)
			return false;
		//dpl("read: text=" + md.text);
		try { readModuleConfig(md.text, mc); }
		catch (Throwable t) {
			dpl("error reading module " + modName + ", " +
				"got error " + t.toString() +
				" error message :" + t.getMessage());
		}
		return true;
	}
	/**
	 * Stores a module's configuration into this object.
	 * @param	modName	the name
	 * @param	m	the ModuleConfig object to be written
	 */
	public void write(String modName, ModuleConfig m) {
		ModDescr md = (ModDescr) mods.get(modName);
		if (md == null)	return;
		String s = ModConfigToString(m);
		//dpl("write: m=" + m);
		//dpl("write: modName="+modName + ",string=" + squote(s));
		md.text = s;
	}
	
	/**
	 * Gets the name of the currently selected module.
	 */
	public String getSelectedModuleName() {
		return header.selected_module;
	}
	
	/**
	 * Sets the name of the currently selected module.
	 */
	public void setSelectedModuleName(String name) {
		header.selected_module = name;
	}
	
	/**
	 * Gets the list of names of packages in which modules
	 * can reside.
	 */
	public String[] getModulePackageNames() {
		return header.module_packages;
	}
	
	/**
	 * Sets the list of names of packages in which modules
	 * can reside.
	 */
	public void setModulePackageNames(String[] names) {
		header.module_packages = names;
	}
	
	// Fills in a ModuleConfig object by reading from a String
	//
	private void readModuleConfig(String s, ModuleConfig m) {
		if (s == null || s.length() == 0) return;
		DataReader dr = new DataReader(new StringReader(s));
		dr.readPublicFields(m);
	}
	
	// Converts the public fields of a ModuleConfig object
	// to a String
	//
	private String ModConfigToString(ModuleConfig m) {
		StringWriter w = new StringWriter();
		DataWriter dw = new DataWriter(w);
		dw.write(m);
		return w.toString();
	}
	
	/**
	 * Duplicates the module entry with the given name.
	 * @param	modName	the name of the existing module
	 * @returns the name of the duplicate, or null if failure.
     */
    public String duplicate(String modName) {
    	ModDescr md = (ModDescr) mods.get(modName);
    	if (md == null)	return null;
    	//dpl("dup: got descriptor");
    	String rv = addMod(md.loadName, md.loadName, md.text);
    	return rv;
    }

	/**
	 * Loads a module, given its name.
	 * Configures the module,
	 * if there is configuration text for it.
	 * This method calls loadKnownModule.
	 *
	 * @param	modName	the module name
	 * @returns the module, or null on failure.
	 * @see #getModNames
	 * @see #loadKnownModule
	 */
    public Module loadModule(String modName) {

    	ModDescr md = (ModDescr) mods.get(modName);
    	if (md == null) {
    		dpl("unknown module name: " + modName);
    		return null;
    	}
    	return loadKnownModule(md);
    }
	
	/**
	 * Loads a module, given a module descriptor.
	 * This method instantiates a module by the
	 * following steps:<br>
	 * <ol>
	 * <li> Load the class. </li>
	 * <li> Invoke the static method
	 * createEmptyConfig() of the module's class.</li>
	 * <li> If createEmptyConfig() returns null,
	 * create the module by invoking the default
	 * constructor, and return. Otherwise,
	 * continue to the next step. <li>
	 * <li> Read the configuration text, if any,
	 * to fill in the config object. </li>
	 * <li> Create the module by invoking a constructor
	 * which takes the configuration object
	 * as a parameter. </li>
	 * </ol>
	 *
	 * @param	md	The module description entry.
     * @returns the module, or null on failure.
     */
    private Module loadKnownModule(ModDescr md) {
    	
    	// load the module class
		//
		String loadName = md.loadName;
		//dpl("loading module " + loadName);
	    Class c = loadModuleClass(loadName);
	    if (c == null) return null;
		String clsname = c.getName();
	
	    // get an empty config object
		//
		String methodName = "createEmptyConfig";
		Method m;
		try {
			m = c.getMethod(methodName, null);
		}
		catch (NoSuchMethodException e) {
			dpl("no method " + clsname + "." + methodName);
			return null;
		}
		if (!Modifier.isStatic(m.getModifiers())) {
			dpl("method " + clsname + "." + methodName +
				" is not static");
			return null;
		}
    	ModuleConfig mc = null;
		try {
			Object obj = m.invoke(null, null);
			if (obj == null) {
				dpl(clsname + "." + methodName + "returned null");
				return null;
			}
			mc = (ModuleConfig) obj; // possible cast exception
		}
		catch (Exception e) {
			dpl("invoke of " + clsname + "." + methodName +
				"failed, exception " + e);
			return null;
		}

		// If there is no config object,
		// just return a module from default constructor.
		//
		Object o = null;
		if (mc == null) {
			try { o = c.newInstance(); }
			catch(Exception e) { 
				dpl("invoked default constructor for " +
				    c.getName() + ",\n" +
				    "exception: " + e);
				return null;
			}
			return (Module) o;
		}

		// fill in the config object
		//
		//dpl("loadKnownModule: text=" + md.text);
		try { readModuleConfig(md.text, mc); }
		catch (Throwable t) {
			dpl("error reading config text for module " + md.name + ",\n" +
				" got error " + t.toString() + "\n" +
				" error message :" + t.getMessage() + "\n" +
				" config text='" + md.text + "'");
		}

		// get a constructor that takes a config object
		// of the type obtained above;
		//
		Constructor con;
		try {
			con = c.getConstructor(new Class[] {mc.getClass()});
		}
		catch (NoSuchMethodException e) {
			dpl("no constructor " + clsname +
				"(" + mc.getClass() + ")");
			return null;
		}

		// invoke the constructor, passing it the config object
		try {
			o = con.newInstance(new Object[] {mc});
		}
		catch (Exception e) {
			dpl("constructor " + clsname + 
				"(" + mc.getClass() + ") failed, exception " + e);
			return null;
		}

		return (Module)o;
		
    }
	
	/**
	 * Finds a module class,
	 * and adds it to the module list.
	 * Calculates a unique module name,
	 * which is returned.
	 *
	 * @param clsname the class name of the module.
	 *  The package list is searched for this class; therefore
	 *  this is not necessarily a full path to the class.
	 *
	 * @returns the module name if successful, null if not.
	 */	
    public String addToModuleList(String clsname) {
	
    	// load the module class
	    Class c = loadModuleClass(clsname);
	    //dpl("loadModuleClass(" + clsname + ") returned " + c);
	    if (c == null) return null;
	
	    // add to the module list
    	String modName = addMod(clsname, clsname, "");
    	
    	return modName;
    }

    /**
     * Attempt to instantiate a class using default contructor
     * @param	c	the class object
     * @returns the object, or null upon failure.
     */
    private Object instantiate(Class c) {

    	Object o = null;
    	try { o = c.newInstance(); }
        catch (Throwable t) {
        	dpl("can't instantiate " + c.getName());
        	return null;
        }
        return o;
    }

    /**
     * Attempt to load a module class, given a class name.
     * This method uses the list of packages
     * (@see #getModulePackageNames).
     * For each package name, this method appends
     * "." + the given name, to construct a full class name,
     * which it attempts to load.  It tries each
     * package name until one succeeds.
     *
     * @param 	clsname	the classname.
     * @returns the class object, if successful,
     * or null if unsuccessful.
     */
    private Class loadModuleClass(String clsname) {
    	String lname;
    	String[] pk = getModulePackageNames();
    	//dpl("loadModuleClass: clsname=" + clsname);
			
    	// note: to be successful, the package name list
    	// must contain at least a zero-length string.
		for (int i = 0; i < pk.length; i++) {
			//dpl("loadModuleClass: pk[i]=" + pk[i]);
			String prefix = pk[i];
			lname = clsname;
			if (prefix.length() > 0) {
				lname = prefix + '.' + clsname;
			}
			Class c = null;
			try { c = Class.forName(lname); }
			// don't use Exception here; could get Error
			catch (Throwable t) {
				//dpl("loadModuleClass: lname="+lname+",t=" + t);
			}
			//dpl("loadModuleClass: c=" + c);
			if (c != null) {
				if (!Module.class.isAssignableFrom(c)) {
					dpl(c.getName() + " not a subclass of Module");
					return null;
				}
				return c;
			}
    	}
    	return null;
    }

	private Conf(boolean junk) { }
	
	// test some of the methods
	//
	public static void main(String[] args) {

    	Conf c = new Conf(true);
    	for (int i=0; i<args.length; i++) {
    		String clsname = args[i];
    		System.out.println(c.loadModuleClass(clsname));
    	}

    	System.out.println("----------------");
    	c.setModulePackageNames(new String[]
    		{ "", "scotthaywood.imageCreator.modules" });
    	System.out.print("packages=");
    	String[] pk = c.getModulePackageNames();
    	for (int i=0; i<pk.length; i++)
    		System.out.print("'" + pk[i] + "' ");
    	System.out.println("");	
    	String mname = c.addToModuleList("SingleColor");
    	System.out.println("mname=" + mname);
    	
    	//String mname = "scotthaywood.imageCreator.modules.SingleColor";
    	//c.addMod(mname, mname, "");
    	ModDescr md = (ModDescr) c.mods.get(mname);
    	Module m = c.loadKnownModule(md);
    	System.out.println("m=" + m);
    }


    /**
     * Writes this configuration to a file.
     * @param	fileName	the file name.
     */
    public void writeFile(String fileName) throws IOException {
		filename = fileName;
		Writer wr = new FileWriter(fileName);
		writeFile(wr);
	}
	/**
     * Writes this configuration to a Writer.
     * @param	wr	the Writer.
     */
    public void writeFile(Writer wr) throws IOException {
    	//dpl("header="+header);
    	String headerText = ModConfigToString(header);
		wr.write(headerText + linesep);
		for (Iterator i=modNames.iterator(); i.hasNext(); ) {
			String name = (String)i.next();
			ModDescr md = (ModDescr) mods.get(name);
			
			// convert the module name to an AsciiIO string,
			// adding quotes and escapes as needed.
			String cname = DataWriter.canonicalString(name);
			String cclsname = null;
			if (!name.equals(md.loadName))
				cclsname = DataWriter.canonicalString(md.loadName);
			wr.write(mkeyword + " " + cname +
				(cclsname == null ? "" :  " " + cclsname) +
				linesep);
			String s = getModText(name);
			wr.write((s==null?"":s) + linesep);
		}
		wr.flush();
		//dpl("exit writeFile");
		
	}
	
	
    private String dplcn;
	private void dpl(String s) {
		if (dplcn == null) {
			String cn = getClass().getName();
			dplcn = cn.substring(cn.lastIndexOf('.')+1);
		}
		System.err.println(dplcn + ": " + s);
	}
	
	private String squote(String s) {
		return "'" + s + "'";
	}
	
	private String myName = "";
	private String mkeyword = "$module";
	private String filename;
    private ArrayList modNames = new ArrayList();
    private Header header = new Header();
    private HashMap mods = new HashMap();
    private final static String linesep = System.getProperty("line.separator");
}
