/*
Image Creator: Creates configurable images
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package scotthaywood.imageCreator;

import java.io.*;
import java.awt.*;
import java.util.*;
import java.awt.image.*;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.*;
import javax.swing.event.*;
import com.sun.image.codec.jpeg.*;
import scotthaywood.imageCreator.ModuleConfig;

/**
 * ImageCreator is an application which displays
 * images and allows them to be written to JPEG files.
 * It allows various "modules" to be loaded dynamically.
 * Each module creates a different type of image.
 * The configuration of the ImageCreator and its
 * modules is stored in a text file.
 *
 * @author Scott Haywood
 * @version $Revision: 1.16 $
 */
public class ImageCreator {

	// class data
	//
	private static java.io.PrintStream se = System.err;
	private static java.io.PrintStream so = System.out;

	// ----------------------------------------------
	
	/**
	 * Runs an ImageCreator application.
	 * @param	args	command strings.
	 */
	public static void main(String[] args) {
	
		String mod_name = null;
		String filename = null;
    	for (int i=0; i<args.length; i++) {
    		if (args[i].equals("-m") && (i+1 < args.length)) {
    			mod_name = args[++i];
    		}
    		else if (args[i].equals("-f") && (i+1 < args.length)) {
    			filename = args[++i];
    		}
    	}
    	
    	if (filename != null && filename.length() > 0) {
    		ImageCreator ic = null;
    		try {
    			ic = new ImageCreator(filename, mod_name);
    		}
    		catch (Throwable t) {
    			perror(t.getMessage());
    			System.exit(-1);
    		}
    		ic.run();
    	} else {
    		perror("expected: -f config_file [ -m module ]");
    		System.exit(-1);
    	}		
    }

    /**
     * Constructs an image creator.
     * @param	cfgFileName	the name of the configuration file.
     * @param	modName	the name of the module to run, or null
     * to run the config file's default module.
     */		
    public ImageCreator(String cfgFileName, String modName) throws Throwable {
    	if (cfgFileName != null) {
    		FileReader fr = null;
    		try { fr = new FileReader(cfgFileName); }
    		catch (Exception e) {
    			perror("Unable to open file " + squote(cfgFileName));
    			throw e;
    		}
    		try { actions.readConfig(fr); }
    		catch (Throwable t) {
    			perror("Error reading file " + squote(cfgFileName) +
    				": " + t.getMessage());
    			throw t;
    		}
    		// handle modName parameter: set the name
    		// of the module to be run.
    		if (modName != null) {
    			conf.setSelectedModuleName(modName);
    		}
    	}
    	
    }


    /**
    	The Actions object contains methods that
    	the UI can call.  Most methods use the <code>Conf</code>
    	object.
    */
    public class Actions {

    	/**
    	 * Exits the application
    	 */
    	public void exit() {
    		System.exit(0);
    	}
    	
    	/**
    	 * Creates a JPEG file.
    	 * @param	m	module of which an image is to be created.
    	 * @param	fos	output stream to which the JPEG file is
    	 * written
    	 */
    	public void createJPEG(ImageModule m, FileOutputStream fos)
    		throws IOException, ImageFormatException
    	{
    		if (m != null) {
    			ImageCreator.createJPEG(m, fos);
    		}
    	}

   		/**
   		 * Saves the configuration of the application.
   		 * @param	fw	FileWriter to which the configuration
   		 * is written.
   		 */
    	public void saveConfig(FileWriter fw)
    		throws IOException
    	{
    		saveToConf();
    		conf.writeFile(fw);
    	}
    	
    	/**
   		 * Reads a new configuration for the application.
   		 * @param	r	Reader from which the configuration
   		 * is read.
   		 */
    	public boolean readConfig(Reader r) throws IOException {
    		Conf cf = null;
    		cf = new Conf(r);
    		if (cf == null)
    			return false;
    		conf = cf;
    		mod = null;
    		return true;
    	}
    	
    	/**
    	 * Gets the names of the modules in this configuration.
    	 * @returns an array of the module names.
    	 */
    	public String[] getModNames() {
    		return conf.getModNames();
    	}
    	
    	/**
    	 * Gets the name of the default module to be run.
    	 * @returns	the module name.
    	 */
    	public String getDefaultModuleName() {
    		String sn = conf.getSelectedModuleName();
    		if (sn != null)
    			return sn;
    		String[] modnames = conf.getModNames();
    		if (modnames != null && modnames.length >0)
    			return modnames[0];
    		else
    			return null;
    	}
    	
    	/**
    	 * Loads a module by name, assuming it is in
    	 * the list of known modules, and configures it if
    	 * there is configuration text.
    	 *
    	 * @param	name	the module name.
    	 *
    	 * @returns	the module, or null on failure.
    	 * @see Conf#loadModule
    	 */
    	public ImageModule loadModule(String name) {
    	
    		saveCurrModuleState();
			Module m = conf.loadModule(name);
			if (m != null) {
    			if (!(m instanceof ImageModule))
					return null;
				conf.setSelectedModuleName(name);
    			if (mod != null)
    				mod.removeChangeListener(modChangeListener);
    			mod = (ImageModule) m;
    			modChanged = false;
    			mod.addChangeListener(modChangeListener);
    		}
			return (ImageModule) m;
    	}
    	
    	/**
    	 * Adds an entry to the list of known module classes.
    	 *
    	 * @see Conf#addToModuleList
		 */	
   		public String addToModuleList(String clsname) {
   		
        	return conf.addToModuleList(clsname);
        }

    	/**
    	 * Duplicates a module, and optionally loads the new one.
    	 * @param	name	the name of the module to be duplicated.
    	 * @param	m	The output module array.
    	 * If its length is > zero,
    	 * stores the new module in it, and
    	 * sets the current module to be the new one.
    	 * @returns the name of the new module, or null on failure.
    	 * @see Conf#duplicate
    	 */
    	public String duplicate(String name, ImageModule[] m) {
    		
    		saveCurrModuleState();
			String newname = conf.duplicate(name);
    		if (newname == null) return null;
    		if (m != null && m.length > 0) {
    			ImageModule im = loadModule(newname);
    			if (im == null) return null;
    			conf.setSelectedModuleName(newname);
    			mod = m[0] = im;
    		}
    		return newname;
    	}
    	
    } // end of Actions


    // instance data
    //

	private String myName = getClass().getName();
    private Actions actions = new Actions();
    private ImageCreatorUI ui = new ImageCreatorUI(actions);
    private Conf conf;
    private ImageModule mod;
    private boolean modChanged;
    private ChangeListener modChangeListener = new ChangeListener() {
    	public void stateChanged(ChangeEvent e) {
    		if (e.getSource() == mod) {
    			modChanged = true;
    		}
    	}
    };

    //
    private String dplcn;
	private void dpl(String s) {
		if (dplcn == null) {
			String cn = getClass().getName();
			dplcn = cn.substring(cn.lastIndexOf('.')+1);
		}
		se.println(dplcn + ": " + s);
	}
   	
    private static void perror(String s) {
    	se.println(s);
    }

    private static String squote(String s) {
    	return "'" + s + "'";
    }


    private static void createJPEG(ImageModule module, OutputStream os)
    	throws IOException, ImageFormatException
    {
		// create BufferedImage and draw to it
		//
		BufferedImage bi = (BufferedImage) module.createImage(
			module.getWidth(),
			module.getHeight());
		Graphics2D g = bi.createGraphics();
		module.paintImage(g);
		
		// encode the BufferedImage
		//
		JPEGImageEncoder jie = JPEGCodec.createJPEGEncoder(os);
		JPEGEncodeParam jep = jie.getDefaultJPEGEncodeParam(bi);
		//jep.setQuality(1.0f, false); // max quality
		jep.setQuality(1.0f, true); // max quality
		jie.setJPEGEncodeParam(jep);
		jie.encode(bi); // might throw exception
    }


    // Save the current state to the Conf object.
    // Note that the indicator of which module
    // is selected is already stored in the Conf object.
    private void saveToConf() {
		//dpl("selected mod=" + conf.getSelectedModuleName());
		
		// save the state of the current module
		// if it has changed.
		saveCurrModuleState();
	}
	
	/**
     * Saves the state of the current module.
     */
    private void saveCurrModuleState() {
    	
    	String sav = conf.getSelectedModuleName();
		if (mod != null && sav != null && modChanged) {
			//dpl("saving the state of module " + sav);
			conf.write(sav, mod.getConfig());
		}
		else {
			//dpl("not saving the state of current module");
			//dpl("(sav="+sav+",modChanged="+modChanged+")");
		}
    }
		
	//
	private void run() {
		ui.run();
	}
}
