/*
Image Creator: Creates configurable images
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package scotthaywood.imageCreator;

import java.io.*;
import java.awt.*;
import java.util.*;
import java.awt.image.*;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import com.sun.image.codec.jpeg.*;
import scotthaywood.ui.*;
import scotthaywood.imageCreator.*;

/**
 * Provides an graphical user interface (GUI) for the
 * ImageCreator application.
 * The application has a list of modules (each one
 * displays a different image).  The "current" module
 * is the one being displayed.
 * <p>
 * The run() of this class method opens two windows:
 * One displays the current module. The other contains GUI
 * components (tools) for selecting a new current module,
 * modifying the configuration of the current module,
 * creating a JPEG file, saving the configuration of 
 * all the modules to a file, loading a new configuration, etc.
 * <p>
 * To do: 
 * <br>
 * (1) Add an warning dialog, which would be invoked
 * when the configuration has changed but not saved to a file.
 * <br>
 * (2) Add an "undo" capability, so that changes to the
 * configuration can be undone.  This would require
 * each module to register undoable changes.
 * <p>
 * Note that a module's size is
 * part of its configuration, and
 * when the size of the current module changes (as when the
 * window containing it is resized) its configuration is changed.
 * 
 * 
 *
 * @author Scott Haywood
 * @version $Revision: 1.19 $
 */
class ImageCreatorUI {

    // ----------------------------------------------
	// class data
	//
	static java.io.PrintStream se = System.err;
	static java.io.PrintStream so = System.out;

    // ----------------------------------------------

    private String dplcn;
	private void dpl(String s) {
		if (dplcn == null) {
			String cn = getClass().getName();
			dplcn = cn.substring(cn.lastIndexOf('.')+1);
		}
		se.println(dplcn + ": " + s);
	}

	public final static int PIXELS_PER_INCH =
		Toolkit.getDefaultToolkit().getScreenResolution();
	
	public static double inch(double x) {
		return x * PIXELS_PER_INCH;
	}

	
    // ----------------------------------------------
	
    // instance data
    //
	private ImageCreator.Actions ica; // always non-null; see constructor
    private JFrame toolFrame;
	private JFrame imageFrame;
	private static class MyLabel extends JLabel {
		public MyLabel(String s) {
			super(s, SwingConstants.CENTER);
		}
		public void sayNow(String s) {
			setText(s);
			paintImmediately(0,0,getWidth(),getHeight());
		}
	}
	private MyLabel message = new MyLabel("Hello");
	private JLabel imageSizeMessage = new JLabel();
	private JButton quit = new JButton("quit");
	private JButton jpeg = new JButton("Create JPEG file");
	private JButton saveButton = new JButton("save configuration");
	private JButton openButton = new JButton("open config. file");
	private ModChooser modChooser = new ModChooser();
	
	private class ModChooser extends JComboBox /* implements MenuElement */{
		
		public ModChooser() {
			/* [ display is too weird with this border ]
			setBorder(new TitledBorder("Select Module"));
			*/
			setEditable(true);
		}
		private ActionListener al = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				Object obj = getSelectedItem();
				Class cls = obj.getClass();
				String loadname = null;
				//dpl("item is '" + obj + "', type: " + cls);
				if (cls == String.class) {
					// The selected item might be a shorter
					// version of the module name.
					// Get the full name.
					loadname = (String) names.get(obj);
					
					// If the user types in a new name,
					// it will not be in the module list.
					// So, attempt to add it.
					if (loadname == null) {
						loadname = ica.addToModuleList((String) obj);
						if (loadname == null) {
							message.setText("can't find: " + obj);
							return;
						}
					}
						
					//dpl("loading " + loadname);
					loadModule2(loadname);
					imageFrame.toFront();
				}
			}
		};
		private void setListener() { addActionListener(al); }
		private void unsetListener() { removeActionListener(al); }
		
		// set of names to be displayed,
		// mapped to the full module names.
		private HashMap names = new HashMap(4);
		
		public void setItems(String[] a, String selected) {
			unsetListener();
			String found = null;
			removeAllItems();
			names.clear();
			for (int i=0; i< a.length; i++) {
				String nm = a[i];
				String displayName = nm;
				
				// Display only the last part of the module name
				// (the part after the last '.'), if it's unique.
				// Otherwise, display the whole name.
				String suffix = nm.substring(nm.lastIndexOf('.')+1);
				if (names.containsKey(suffix)) {
					if (names.containsKey(nm)) {
						perror("duplicate module name: " + nm);
					}
				} else {
					displayName = suffix;
				}
				
				// Remember all the names to be displayed,
				// and their associated full names.
				names.put(displayName, nm);
				
				addItem(displayName);
				if (found == null && nm.equals(selected)) {
					found = displayName;
				}
			}
			if (found != null) setSelectedItem(found);
			//perror("setItems: count=" + getItemCount());
			setListener();
		}
		
		/*
		// implementing MenuElement:
		public void processMouseEvent(MouseEvent e,
		                              MenuElement path[],
		                              MenuSelectionManager manager){}
		public void processKeyEvent(KeyEvent e,
		                              MenuElement path[],
		                              MenuSelectionManager manager){}
		
		public void menuSelectionChanged(boolean isIncluded) {}
		public MenuElement[] getSubElements() {
			return new MenuElement[0];
		}
		public Component getComponent() { return this; }
		*/
		
	};
			
	private String myName = getClass().getName();
	private String myTitle;
	private String defaultDirectory;
    private ImageModule mod; // the currently selected module
    private String modName; // name of the currently selected module
    private JPanel modToolPanelContainer = new JPanel();
    private ComponentAdapter resizeListener = new ComponentAdapter() {
		public void componentResized(ComponentEvent e) {
			setImageSizeMessage();
		}
	};
	private ColorIcon blankIcon = new ColorIcon(200, 200, Color.black);
	private JLabel blankImage = new JLabel(blankIcon);
	private Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();

    private Component modToolPanel;

	// -----------------------------------------------------------------
    // Actions

	// Exit the application
	//
	private AbstractAction quitAction = new AbstractAction(quit.getText()) {
		public void actionPerformed(ActionEvent e) {
				ica.exit();
		}
	};

	// create a JPEG file containing an image of the current module
	//
	private AbstractAction jpegAction = new AbstractAction(jpeg.getText()) {
		public void actionPerformed(ActionEvent e) {
			if (mod == null) return;
			JFileChooser fc = getFileChooser();
			int opt = fc.showSaveDialog(toolFrame);
			if (opt != JFileChooser.APPROVE_OPTION) {
				return;
			}
			File file = fc.getSelectedFile();
			FileOutputStream fos = null;
			try {
				fos = new FileOutputStream(file); 
			}
			catch (Exception fose) {
				message.setText("Could not open " +
					squote(file.getPath()) + " for writing");
				return;
			}
			message.setText("opened  "+ squote(file.getPath()));
			
			try {
				ica.createJPEG(mod, fos);
			}
			catch (IOException ioe) {
				message.setText(
					"i/o error when creating the JPEG file");
				return;
			}
			catch (ImageFormatException ife) {
				message.setText(
					"ImageFormatException when creating the JPEG file");
				return;
			}
			message.setText("JPEG file '" + 
					file.getName() + "' created: " +
					file.length() + " bytes");
			return;
		}
	};

	// save the configuration of the application to a file
	//
	private AbstractAction saveAction = 
	  new AbstractAction(saveButton.getText()) {
		public void actionPerformed(ActionEvent e) {

			JFileChooser fc = getFileChooser();
			int opt = fc.showSaveDialog(toolFrame);
			if (opt != JFileChooser.APPROVE_OPTION) {
				return;
			}
			File file = fc.getSelectedFile();
			//perror("got file " + file.getPath());
			FileWriter fw = null;
			try {
				fw = new FileWriter(file);
			}
			catch (Exception fe) {
				message.setText("can't open " +
					squote(file.getPath()) + " for writing");
				return;
			}
			message.setText("opened " + file.getPath());
			try {
				ica.saveConfig(fw);
			}
			catch (IOException ioe) {
				message.setText(
					"i/o error when saving the configuration file");
				return;
			}
			message.setText("file '" +
					file.getPath() + "' created: " +
					file.length() + " bytes");
		}
	};
	
	// read a the configuration of the application from a file
	//
	private AbstractAction readAction  = 
	  new AbstractAction(openButton.getText()) {
		public void actionPerformed(ActionEvent e) {

			JFileChooser fc = getFileChooser();
			int opt = fc.showOpenDialog(toolFrame);
			if (opt != JFileChooser.APPROVE_OPTION) {
				return;
			}
			File file = fc.getSelectedFile();
			//perror("got file " + file.getPath());
			FileReader fr = null;
			try { fr = new FileReader(file); }
			catch (Exception fe) {
				message.setText("can't open " +
					squote(file.getPath()) + " for reading");
				return;
			}
			message.setText("opened " + file.getPath());
			boolean flg = false;
			try { flg = ica.readConfig(fr); }
			catch (Throwable t) {
				message.setText(
					"i/o error when reading the configuration file");
				return;
			}
			finally { 
				//perror("flg=" + flg + ", read " + (flg?"succeeded":"failed"));
			}
			message.setText("file '" +
					file.getPath() + "' read: " +
					file.length() + " bytes");
					
			getDefaultModule();
		}
	};
	
	// Duplicate the current module.
	// This allows the user to have two different configurations
	// of the same module.
	//
	private AbstractAction dupAction = new AbstractAction("clone image") {
		public void actionPerformed(ActionEvent e) {
			perror("Enter dup");
			String s = ica.duplicate(modName, null);
			if (s != null) {
				loadModule2(s);
			}
			perror("Exit dup");
			
		}
	};
	
	// end of Actions
	// -----------------------------------------------------------------
	
	// Sets the directory to be used by getFileChooser()
	private void setDefaultDirectory(String dir) {
		defaultDirectory = dir;	
	}
	
	private JFileChooser getFileChooser() {
		String cwd;
		JFileChooser fc;
		if (true) {
			fc = new JFileChooser(defaultDirectory);
		} else if (false) {
			cwd = System.getProperty("user.dir");
			fc = new JFileChooser(cwd);
		} else {
			fc = new JFileChooser(".");
		}
		fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
		return fc;	
	}
	
	private void updateModuleList() {
		String[] names = ica.getModNames();
		modChooser.setItems(names, modName);
	}
	
	
	// Attempt to load a module, given its name.
	private ImageModule loadModule(String name) {
	
		message.sayNow("loading " + squote(suffix(name)) + "...");
		//dpl("loading");
		ImageModule m = ica.loadModule(name);
		message.setText((m == null) ?
					    "can't load module " + squote(name) :
					    "loaded " + squote(suffix(name)));
		return m;
	}
	
	private String suffix(String s) {
		return s.substring(s.lastIndexOf('.')+1);
	}
	
	// Attempt to load a module, given its name.
	// If load succeeds, set the current module
	// and update the module list.
	private void loadModule2(String name) {
		ImageModule m = loadModule(name);
		if (m != null) {
			setModule(name, m);
			updateModuleList();
		}
	}
	
	// get the default module name, attempt to load.
	// Set the current module (to null if load fails).
	private void getDefaultModule() {
		String name = ica.getDefaultModuleName();
		ImageModule m = null;
		if (name != null)
			m = loadModule(name);
		setModule(name, m);
		updateModuleList();
	}
    //

    static void perror(String s) {
    	se.println(s);
    }

    static String squote(String s) {
    	return "'" + s + "'";
    }

	// Format a message that displays the image size.
	//
	private String formatImageSize(Dimension d) {
		int len=14;
		StringBuffer sb = new StringBuffer(
			"w=" + d.width + ", h=" + d.height);
		while (sb.length() < len) sb.append(' ');
		return sb.substring(0,len);
	}
	
	// Display the size of the current module.
	//
	private void setImageSizeMessage() {
		imageSizeMessage.setText(
			mod == null ? "" : formatImageSize(mod.getSize()));
	}
	
	// Stop displaying the current module's tool panel.
	//
	private void removeModToolPanel() {
		modToolPanelContainer.removeAll();
	}
	
	// Display the current module's tool panel.
	//
	private void addModToolPanel() {
		if (mod != null &&
		    (modToolPanel = mod.getToolPanel()) != null)
		{
			modToolPanelContainer.add(modToolPanel);
		}
	}
	
	// Set the current module,
	// and change the display accordingly.
	// Listen for a resize of the new module, and not the old.
	//
	private void setModule(String name, ImageModule m) {
		
		modName = name;
		
		// clean up old module
		if (mod != null) {
			mod.removeComponentListener(resizeListener);
		}
		removeModToolPanel();
		
		// setup for new module
		mod = m;
		setImageSizeMessage();
		
		if (mod == null) {
			jpeg.setEnabled(false);
			jpegAction.setEnabled(false);
			imageFrame.setContentPane(blankImage);
			imageFrame.setTitle(myTitle);
		}
		else {
			jpeg.setEnabled(true);
			jpegAction.setEnabled(true);
		    mod.addComponentListener(resizeListener);
		    addModToolPanel();
			/*
		    Container c = imageFrame.getContentPane();
		    c.removeAll();
			c.setSize(mod.getPreferredSize());
		    c.setLayout(new BorderLayout());
		    c.add(mod.getDisplayPanel(), BorderLayout.CENTER);
			*/
			imageFrame.setContentPane(mod.getDisplayPanel());
	    	imageFrame.setTitle(myTitle + ": " + modName);
	    }
	    /*
	    imageFrame.pack();
	    if (mod != null) {
			so.println("after pack(): panel width,height="
				+mod.getWidth() + ","
				+mod.getHeight());
		}
		*/
		
		setFrameLocations();
		
	    //toolFrame.grabFocus();
		
	}
	
	// Set the location of the display frames
	//
	private void setFrameLocations() {
		imageFrame.pack();
	    toolFrame.pack();
		int h1 = imageFrame.getHeight();
		int h2 = toolFrame.getHeight();
		int w1 = imageFrame.getWidth();
		int w2 = toolFrame.getWidth();
		int x1, x2, y1, y2;

		//dpl("h1=" + h1 + ",h2=" + h2 + ",w1=" + w1 + ",w2=" + w2);
		if (h1 + h2 <= screenSize.height) {
			x1 = (screenSize.width - w1)/2;
			x2 = (screenSize.width - w2)/2;
			y1 = (screenSize.height - h1 - h2)/2;
			y2 = y1 + h1;
		}
		else if (w1 + w2 <= screenSize.width) {
			x1 = (screenSize.width - w1 - w2)/2;
			x2 = x1 + w1;
			y1 = (screenSize.height - h1)/2;
			if (y1 < 0) y1 = 0;
			y2 = (screenSize.height - h2)/2;
			if (y2 < 0) y2 = 0;
		} else {
			x1 = y1 = 0;
			final int min_offset = 18;
			x2 = (min_offset + w2 > screenSize.width)
				?  min_offset : screenSize.width - w2;
			y2 = (min_offset + h2 > screenSize.height)
				? min_offset : screenSize.height - h2;
		}
		//dpl("x1=" + x1 + ",x2=" + x2 + ",y1=" + y1 + ",y2=" + y2);
        imageFrame.setLocation(x1, y1);
        toolFrame.setLocation(x2, y2);
	}
	
	
	/**
	 * Runs the user interface.
	 */
	public void run() {
	
		getDefaultModule();
		
		//imageFrame.pack();
		Thread thread = new Thread() {
			public void run() {
				// may not need a listener here
				imageFrame.addWindowListener(new WindowAdapter() {
					public void windowClosing(WindowEvent e) {
            			throw new RuntimeException("close");
            		}
				});
				try { imageFrame.show(); }
				catch (Throwable t) {
					perror(myName+".run: "+t.getMessage());
				}
			}
		};
		thread.start();
		toolFrame.show();
	}
	
	
	// ----------------------------------------------
	// constructors

    /**
	 * Constructs a user interface for the ImageCreator application.
	 */
	public ImageCreatorUI(ImageCreator.Actions icActions){
    	this(icActions, null);
    }

    /**
	 * Constructs a user interface for the ImageCreator application.
	 */
	public ImageCreatorUI(ImageCreator.Actions icActions,
	                      String title)
	{
		// Remember the current working directory
		setDefaultDirectory(System.getProperty("user.dir"));
		
		ica = icActions;
    	if (ica == null)
    		throw new Error(myName + ": null ImageCreator.Actions");

    	myTitle = title;
    	if (myTitle == null) {
    		myTitle = myName.substring(myName.lastIndexOf('.')+1);
    		myTitle = myTitle.substring(0,
    			myTitle.endsWith("UI") ?
    			myTitle.lastIndexOf("UI") :
    			myTitle.length());
    	}
		imageFrame = new JFrame(myTitle);

		// attempt transparent background 
		//imageFrame.setBackground(new Color(0,true));

		toolFrame = new JFrame(myTitle + " controls");
        WindowAdapter wa = new WindowAdapter() {
        	public void windowClosing(WindowEvent e) {
            	imageFrame.setVisible(false);
            	imageFrame.dispose();
            	toolFrame.setVisible(false);
            	toolFrame.dispose();
            	ica.exit();
            }
        };
        toolFrame.addWindowListener(wa);
        imageFrame.addWindowListener(wa);

		// create controls
		//
		JMenu fileMenu = new JMenu("File");
		fileMenu.add(readAction);
		fileMenu.add(saveAction);
		fileMenu.add(jpegAction);
		fileMenu.add(dupAction);
		fileMenu.add(quitAction);
		
		//JMenu modMenu = new JMenu("Module");
		//modMenu.add(modChooser);
		
		JMenuBar menuBar = new JMenuBar();
		menuBar.add(fileMenu);
		//menuBar.add(modMenu);
		toolFrame.setJMenuBar(menuBar);
		
		quit.addActionListener(quitAction);
		jpeg.addActionListener(jpegAction);
		saveButton.addActionListener(saveAction);
		
		Box myModControls = Box.createHorizontalBox();
		myModControls.add(Box.createHorizontalGlue());
	    //int strut_size = (int)Math.round(inch(0.4));
		//myModControls.add(Box.createHorizontalStrut(strut_size));
		myModControls.add(imageSizeMessage);
		myModControls.add(jpeg);
		//myModControls.add(Box.createHorizontalStrut(strut_size));
		myModControls.add(Box.createHorizontalGlue());
		
		Box progControls = Box.createHorizontalBox();
		progControls.add(Box.createHorizontalGlue());
		progControls.add(saveButton);
		progControls.add(quit);
		progControls.add(Box.createHorizontalGlue());
		
        toolFrame.getContentPane().setLayout(new BorderLayout());
        toolFrame.getContentPane().add(message, "North");

        // reserve a place for the module's tool panel
        modToolPanelContainer = new JPanel();
        modToolPanelContainer.setBorder(
        	BorderFactory.createEmptyBorder(4,4,4,4));
        toolFrame.getContentPane().add(
        	modToolPanelContainer, "Center");

        // create a vertical box with a border.
        // note that Box can't take a border.	
        JPanel controlPanel = new JPanel();
        int horizBorder = 8;
        controlPanel.setBorder(BorderFactory.createEmptyBorder(
        	4,horizBorder,4,horizBorder));
        controlPanel.setLayout(
        	new BoxLayout(controlPanel, BoxLayout.Y_AXIS));
        Box lbox = Box.createHorizontalBox();
        lbox.add(new JLabel("Select image:", SwingConstants.LEFT));
        lbox.add(Box.createHorizontalGlue());
        controlPanel.add(lbox);
        controlPanel.add(modChooser);
        controlPanel.add(myModControls);
        controlPanel.add(progControls);
		toolFrame.getContentPane().add(controlPanel, "South");

		setModule(null, null);

	}
}
