/*
Image Creator: Creates configurable images
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package scotthaywood.imageCreator;

import java.awt.Graphics;
import java.awt.Dimension;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import javax.swing.JComponent;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import scotthaywood.imageCreator.*;

/**
	An <code>ImageModule</code> is a type of <code>Module</code>,
	which (1) displays an image, and (2) provides a user interface
	for modifying the image, (3) automatically changes its configuration
	object when it is resized, and (4) repaints itself whenever
	it fires a "state changed" event.
	<p>
	<code>ImageModule</code> must be subclassed; it provides no
	public constructor.  Subclasses normally override the following
	methods:
	<tt><ul>
	<li>getToolPanel</li>
	<li>getConfig</li>
	<li>setConfig</li>
	</ul></tt>
	A subclass might also need to override:
	<tt><ul>
	<li>getDisplayPanel</li>
	<li>paintImage</li>
	</ul></tt>
	
	@see Module
	@see #getToolPanel
	@see #getConfig
	@see #setConfig
	@see #getDisplayPanel
	@see #paintImage
 *
 * @author Scott Haywood
 * @version $Revision: 1.7 $
*/
public class ImageModule extends Module {

	private String myName = getClass().getName();
	
	protected ImageModule() {
	
		/*
			When this component is resized,
			store the new size in the ImageModuleConfig member,
			and generate a ChangeEvent.
			
		*/
		//perror("Enter ImageModule()");
	    addComponentListener(new ComponentAdapter() {
	    	public void componentResized(ComponentEvent e) {
	    		//perror(myName + ": resized");
	    		ModuleConfig mc = getConfig();
	    		if ((mc != null) &&
	    		    (mc instanceof ImageModuleConfig)) {
	    		    ((ImageModuleConfig)mc).setSize(getSize());
	    		    fireStateChanged();
	    		}
	    	}
	    });
	
	    /*
	    *  Invoke a default piece of code
	    *  whenever the module calls
	    *  fireStateChanged().  To alter this behavior,
	    *  do:
	    *		removeChangeListener(getChangeListener());
	    *		addChangeListener( .... );
	    */
	    addChangeListener(getChangeListener());
	}

	/*
	*	Define the default ChangeListener.
	*	It repaints the module.
	*/
	private ChangeListener modChangeListener =
		new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				repaint();
			}
		};
	protected ChangeListener getChangeListener() {
		return modChangeListener;
	}
	
	/**
	 * Gets the tool panel, which provides a user interface
	 * for modifying the image module.
	 * Subclasses normally override this method.
	 * @returns the tool panel.
	 */
	public JComponent getToolPanel() {return null;}

	/** 
	 * Gets the configuration object.
	 * Subclasses should override getConfig() to
	 * return an ImageModuleConfig object.
	 * (ImageModuleConfig is a subclass of ModuleConfig).
	 * The public fields of the configuration object
	 * are used to save state of the module.
	 * If getConfig() returns null, it means the
	 * module never has a configuration object.
	 *
	 * @returns the configuration object.
	 * @see #setConfig
	 * @see ModuleConfig
	 * @see ImageModuleConfig
	 */
	public ModuleConfig getConfig() {return null;}
	
	/**
	 * Sets the configuration object.
	 * Subclasses should override this method.
	 * It is called when the state of the module is
	 * restored.
	 * 
	 * @param	mc	the new configuration object.
	 * @see #getConfig
	 */
	public void setConfig(ModuleConfig mc) {}
	
	/**
	 * Paints the image for this module. This method is
	 * called when a JPEG image is created.
	 * By default, this calls the module's <code>paint</code>
	 * method.  A subclass may override this, for example to
	 * have it call <code>paintComponent</code> instead.
	 *
	 * @param	g	the graphics object used for painting the image.
	 */
	public void paintImage(Graphics g) {paint(g);}
	

	/**
	 * Returns the component which displays the module's
	 * image as part of the user interface.
	 * By default it returns this ImageModule.
	 * A subclass may override it to return a component
	 * which contains this ImageModule.
	 *
	 * @returns	the component to be displayed.
	 * @see ImageCreator#createJPEG
	 * @see ImageCreatorUI
	 */
	public JComponent getDisplayPanel() { return this; }

}
