/*
Image Creator: Creates configurable images
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package scotthaywood.imageCreator;

import java.awt.Dimension;
import scotthaywood.asciiIO.BasicTypeTranslator;
import scotthaywood.asciiIO.TypeTranslator;
import scotthaywood.imageCreator.*;

/**
	This is a subclass of ModuleConfig.
	It has members for recording the dimensions of the module.
	<p>
	This class provides "translation"
	for two types: SimpleFont and SimpleColor.
	Subclasses of ImageModuleConfig
	can have public fields of thoses types, and be
	written/read as an ascii stream (by the AsciiIO classes),
	without additional coding in the subclass.
	@see ImageModule
 *
 * @author Scott Haywood
 * @version $Revision: 1.6 $
*/
abstract public class ImageModuleConfig extends ModuleConfig {

	/**
	 * The image width in pixels.
	 * It is made public so that it can be written/read by
	 * the AsciiIO classes.
	 */
	public int width;

	/**
	 * The image height in pixels.
	 * It is made public so that it can be written/read by
	 * the AsciiIO classes.
	 */
	public int height;
	
	/**
	 * Returns a string representation of this object.
	 */
	public String toString() {
		String base = getClass().getName();
		base = base.substring(base.lastIndexOf('.')+1);
		return base+"[width="+width+",height="+height+"]";
	}

	/**
	 * Sets the image width and height.
	 * @param w	the image width in pixels
	 * @param h	the image height in pixels
	 * @see #getSize
	 */
    public void setSize(int w, int h) {
    	width = w;
    	height = h;
    }
    /**
	 * Sets the image width and height.
	 * @param d	the Dimension containing the new width and height
	 * @see #getSize
	 */
    public void setSize(Dimension d) {
    	setSize(d.width, d.height);
    }

    /**
	 * Gets the image width and height.
	 * @returns a Dimension containing the width and height
	 * @see #setSize(int, int)
	 * @see #setSize(Dimension)
	 * @see #ensureMinSize
	 */
    public Dimension getSize() {
    	return new Dimension(width, height);
    }

    /**
	 * Sets the image width and height if either is less than
	 * a given minimum.
	 * @param d	the Dimension containing the desired minimum
	 *  width and height
	 * @see #getSize
	 */
    public void ensureMinSize(Dimension d) {
    	if (width < d.width) width = d.width;
        if (height < d.height) height = d.height;
    }

    private static BasicTypeTranslator fontTrans =
		new BasicTypeTranslator(new SimpleFontSurrogate());
	private static SimpleColorTranslator colorTrans =
		new SimpleColorTranslator();
	
	/**
	 * Constructs an ImageModuleConfig object.
	 */
    public ImageModuleConfig() {
	
		// Provide translation for SimpleFont and SimpleColor.
    	// See also Magnifier, which has a built-in TypeTranslator.
		//
		addTypeTranslators(new TypeTranslator[] {
			fontTrans,
			colorTrans
		});
	}
}
