/*
Image Creator: Creates configurable images
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


package scotthaywood.imageCreator;

import java.util.*;
import java.awt.*;
import java.awt.font.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import scotthaywood.asciiIO.*;
import scotthaywood.imageCreator.*;

/**
 * A <code>Magnifier</code> displays another component
 * with an image of a magnifying glass over it, and part
 * of the component magnified (scaled) within the "lens"
 * part of the magnifying glass.
 * It allows the mag. glass to be moved and its size and
 * power to be changed.
 * For convenience, this is a subclass of ImageModule.
 * It is intended to be used by other ImageModule's,
 * rather than to be given directly as a module
 * to the ImageCreator application.
 *
 * @author Scott Haywood
 * @version $Revision: 1.6 $
 */
public class Magnifier extends ImageModule implements AsciiIO.Translatable {

    private String dplcn;
	private void dpl(String s) {
		if (dplcn == null) {
			String cn = getClass().getName();
			dplcn = cn.substring(cn.lastIndexOf('.')+1);
		}
		System.err.println(dplcn + ": " + s);
	}

	private int round(double d) {
   		return (int)Math.round(d);
   	}

	private static Dimension defaultSize = new Dimension(200,200);
	private static Dimension minSize = new Dimension(10,10);
	
	/**
	 * This defines the state of the magnifier
	 */
	public static class Config extends ImageModuleConfig {
	
		/**
		 * Distance of the center of the magnifier circle
		 * from the left edge of the image,
		 * as a fraction of image width.
		 */
		public float xfactor = 0.5f;
    	
    	/**
		 * Distance of the center of the magnifier circle
		 * from the top of the image,
		 * as a fraction of image height.
		 */
		public float yfactor = 0.6f;
		
    	/**
		 * Radius of the magnifier "rim",
		 * as a fraction of the lesser of image height and width.
		 */
		public float radius_factor = 0.25f;
    	
    	/**
    	 * Thickness of the "rim" of the magnifying glass,
    	 * as a fraction of the radius.
    	 */
		public float rim_thickness = 0.04f;
		
		/**
    	 * Thickness of the "handle" of the magnifying glass,
    	 * as a fraction of the radius.
    	 */
		public float handle_thickness = 0.25f;
		
		/**
    	 * Length of the "handle" of the magnifying glass,
    	 * as a fraction of the radius.
    	 */
		public float handle_length = 3f;
		
		/**
    	 * The power of magnification of the image within
    	 * the circle.
    	 */
		public float magnification = 2.5f;

		public Config() {
			super.setSize(defaultSize);
		}
	}

	private Config cfg;
	
	// the target component
	private Component view;
	
	// the tooltip text
	protected static String tip = "The magnifier can be moved";

	/**
	 * Set the target component
	 * @param	c	the new target component
	 */
	public void setView(Component c) {
		view = c;
	}
	
	/**
	 * Gets the target component
	 * @returns	the new target component
	 */
	public Component getView() {
		return view;
	}

	// ------------------------------------------
	// Constructors required of an ImageModule
	
	/**
	 * Constructs a magnifier.
	*/
	public Magnifier() {
		cfg = new Config();
		init();
	}

	private Magnifier(Config c) {
		cfg = c;
		init();
	}
    //----------------------------------------------

    // cursor used when moving the mag. glass
	protected Cursor moveCursor=Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR);

	// mouse event handler for moving the mag. glass.
	//
	private MouseInputAdapter mia = new MouseInputAdapter() {

		Point pressed = null;
		Point pCenter = null;
		Cursor oldCursor = null;

		public void mousePressed(MouseEvent e) {
			Point p = e.getPoint();
			if (all != null && all.contains(p)) {
				//dpl("yes");
				pressed = p;
				int w = getWidth();
				int h = getHeight();
				int x = (int)Math.round(w * cfg.xfactor);
				int y = (int)Math.round(h * cfg.yfactor);
				pCenter = new Point(x,y);
				oldCursor = Magnifier.this.getCursor();
				Magnifier.this.setCursor(moveCursor);
			}
			//else dpl("no");
		}
		public void mouseDragged(MouseEvent e) {
			if (pressed == null) return; // should not happen
			Point p = pressed;
			Point q = e.getPoint();
			//dpl("p="+p+", q="+q);

			int w = getWidth();
			int h = getHeight();
			int x = pCenter.x + q.x - p.x;
			int y = pCenter.y + q.y - p.y;
			boolean changed = false;
			if (x >=0 && x < w ) {
				cfg.xfactor = (float)x/ w;
				changed = true;
			}
			if (y >=0 && y < h) {
				cfg.yfactor = (float)y / h;
				changed = true;
			}
			if (changed) {
				//dpl("mia: xfactor="+cfg.xfactor+", yfactor="+cfg.yfactor); 
				fireStateChanged();
			}
		}
		public void mouseReleased(MouseEvent e) {
			pressed = null;
			Magnifier.this.setCursor(oldCursor);
		}
		
	};

	// initialize this object. see the constructors.
	//
	private void init() {

		setOpaque(false);
		cfg.ensureMinSize(minSize);
		Dimension d = cfg.getSize();
		setPreferredSize(d);
		setToolTipText(tip);
		addMouseListener(mia);
		addMouseMotionListener(mia);
	}

	/** A <code>MagTranslator</code> is used when saving/restoring a
	 * <code>Magnifier</code> to/from a stream of ascii text.
	 */
	public static class MagTranslator implements TypeTranslator {
		/*
		* methods implementing TypeTranslator
		*/
		public Class[] getTargetTypes() {
			return new Class[] {Magnifier.class};
		}
		public Class getSurrogateType(Class targetType) {
			return Config.class;
		}
		public Object createSurrogate(Object target) {
			//if (target.getClass() != Magnifier.class)
			//	throw new Error("wrong target type: " + target.getType());
			Magnifier m = (Magnifier) target;
			return m.cfg;
		}
		public Object createTarget(Class cls, Object surrogate) {
			if (cls != Magnifier.class) return null;
			Config c = (Config) surrogate;
			return new Magnifier(c);
		}
	}

	private static MagTranslator translator;

	/**
	 * Get the <code>TypeTranslator</code> for type
	 * <code>Magnifier</code>.
	 * A method like this is required at runtime for any class
	 * implementing <code>AsciiIO.Translatable</code>.
	 *
	 * @returns the TypeTranslator for class Magnifier
	 */
	public static TypeTranslator getTypeTranslator() {
		if (translator == null)
			translator = new MagTranslator();
		return translator;
	}


	// fields used by paintComponent
	//
	private Ellipse2D.Double outer, inner;
	private Rectangle2D.Double handleRect = new Rectangle2D.Double();
	private GeneralPath all;

	/**
	 * Paints this component.  Overrides JComponent.
	 * @param	graphics	the java.awt.Graphics object.
	 */
	public void paintComponent(Graphics graphics) {
		//dpl("paint: xfactor="+cfg.xfactor+", yfactor="+cfg.yfactor); 
		int w = getWidth();
		int h = getHeight();
		int min = Math.min(w,h);
		int x = (int)Math.round(w * cfg.xfactor);
		int y = (int)Math.round(h * cfg.yfactor);
		float radius = min * cfg.radius_factor;
		float handleThickness = radius * cfg.handle_thickness;
		float handleLen = radius * cfg.handle_length;
		//dpl("x="+x+", y="+y+", radius="+radius);

		Graphics2D g = (Graphics2D) graphics;

		// draw the magnifier rim
		//
		float linew = cfg.rim_thickness * radius;
		outer = new Ellipse2D.Double(
			x - radius - linew/2f, y - radius - linew/2f,
			2 * radius + linew, 2 * radius + linew );
		inner = new Ellipse2D.Double(
			x - radius + linew/2f, y - radius + linew/2f,
			2 * radius - linew, 2 * radius - linew );
		GeneralPath rim = new GeneralPath(outer);
		rim.append(inner, false);
		g.fill(rim);

		// draw the handle
		//
		handleRect.setRect(
			x + radius, y - handleThickness/2f,
			handleLen, handleThickness
			);
		AffineTransform at = new AffineTransform();
		at.rotate(Math.toRadians(45), x, y);
		Shape handle = at.createTransformedShape(handleRect);
		g.fill(handle);

		// remember the entire shape (rim and handle)
		//
		all = new GeneralPath(outer);
		all.append(handle, false);

		// draw the magnified part of the target component
		//
		g.setClip(inner);
		//if (view != null) setBackground(view.getBackground());
		g.clearRect(0,0,w,h);

		float mag = cfg.magnification;
		//dpl("mag="+mag);

		float transX = x * (mag - 1);
		float transY = y * (mag - 1);

		g.translate(-transX, -transY);
		g.scale(mag, mag);
	
		  //dpl("here goes...");
		  if (view != null)
			  view.paint(g);
		  //dpl("...paint done");
		
		g.scale(1/mag, 1/mag);
		g.translate(transX, transY);

	}

	// fields used to create the tool panel
	
	private JPanel toolPanel;
	private int sliderMax = 500;

	private float maxMag = 10f;
	private float minMag = 0.1f;
	private JSlider magSlider;

	private float maxRad = 0.5f;
	private float minRad = 0.01f;
	private JSlider radiusSlider;
	
	
    /**
     * Gets a panel, containing sliders, etc.,
     * used to interactively change
     * this object.
     * @returns the tool panel
     * @see Module
     */
	public JComponent getToolPanel() {

		if (toolPanel != null)
			return toolPanel;
		toolPanel = new JPanel();
		toolPanel.setLayout(new BorderLayout());

		Box box = Box.createVerticalBox();
		toolPanel.add(box, BorderLayout.CENTER);

		magSlider = new JSlider(0,sliderMax, round(
			sliderMax * (cfg.magnification - minMag) / (maxMag-minMag)));
		//dpl("mag slider val="+magSlider.getValue());
		magSlider.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {	
				float v = magSlider.getValue();
				//dpl("event: slider val="+magSlider.getValue());
				float newMag = minMag + (v/sliderMax)*(maxMag-minMag);
				cfg.magnification = newMag;
				//dpl("event: newMag="+newMag);
				fireStateChanged();
			}
		});
		box.add(Box.createVerticalStrut(4));
		box.add(new JLabel("magnification:"));
		box.add(magSlider);

		radiusSlider = new JSlider(0,sliderMax, round(
			sliderMax * (cfg.radius_factor - minRad) / (maxRad-minRad)));
		radiusSlider.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {	
				float v = radiusSlider.getValue();
				float newRad = minRad + (v/sliderMax)*(maxRad-minRad);
				cfg.radius_factor = newRad;
				fireStateChanged();
			}
		});
		box.add(Box.createVerticalStrut(4));
		box.add(new JLabel("magnifier size:"));
		box.add(radiusSlider);

		return toolPanel;
	}
}

