/*
Image Creator: Creates configurable images
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package scotthaywood.imageCreator.modules;

import java.awt.*;
import java.beans.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.awt.font.*;
import javax.swing.*;
import javax.swing.event.*;
import scotthaywood.ui.*;
import scotthaywood.imageCreator.*;

/**
 * GradientText is a type of module for the ImageCreator
 * application.
 * Its image consists of one line of text,
 * painted with a gradient of color,
 * surrounded by an optional border.
 *
 * It provides a user interface for changing
 * the text and its font, selecting the
 * colors and points of the gradient, changing
 * the font, background color,
 * configuring the border, and for selecting
 * to draw the image with or without antialiasing.
 *
 * @author Scott Haywood
 */
public class GradientText extends ImageModule {

    private static GraphicsEnvironment ge =
        GraphicsEnvironment.getLocalGraphicsEnvironment();
   	private static String fontFamilyNames[] = ge.getAvailableFontFamilyNames();
   	private static Font font = new Font("Lucida Sans", Font.PLAIN, 16);

    private String dplcn;
	private void dpl(String s) {
		if (dplcn == null) {
			String cn = getClass().getName();
			dplcn = cn.substring(cn.lastIndexOf('.')+1);
		}
		System.err.println(dplcn + ": " + s);
	}

	static java.io.PrintStream err = System.err;
   	
   	private int round(double d) {
   		return (int)Math.round(d);
   	}
   	
   	private static Dimension defaultSize = new Dimension(400,172);
	private static Dimension minSize = new Dimension(1,1);
	
    /**
	 * The inner class <code>Config</code>
	 * is the type of configuration object for this module type.
	 * Its public fields contain all the information needed
	 * to save and restore a module.
	 */
    public static class Config extends ImageModuleConfig {

    	/**
    	 * Creates a configuration object.
    	 */
    	public Config() {
    		super.setSize(defaultSize);
    	}
    	
    	// default values
    	//
    	static SimpleFont dFont =
    		new SimpleFont("Lucida Sans", Font.PLAIN, 32);
    	static SimpleColor dColor1 = new SimpleColor(32, 128, 64);
    	static SimpleColor dColor2 = new SimpleColor(0,0,153);
    	static String dText = "Hello";
    		
		/**
		 * The text to be displayed.
		 */
		public String text = dText;
		
		/**
		 * The font.
		 * Note: the text is automatically resized.  The font size
		 * is used only for display when selecting a new font.
    	 */
		public SimpleFont font = dFont;
		
		/**
		 * Ending color of the gradient.
		 */
    	public SimpleColor color2 = dColor1;
    	
    	/**
		 * Beginning color of the gradient.
		 */
    	public SimpleColor color1 = dColor2;
    	
    	/**
    	 * Background color
    	 */
    	public SimpleColor bgColor = SimpleColor.white;
    	
    	/**
    	 * Outline color
    	 */
    	public SimpleColor outlineColor =  SimpleColor.black;
    	
    	/**
    	 * Ratio of the outer border size to
    	 * the lesser of the component width and height.
    	 */
    	public float borderFactor1 = 0;

        /**
    	 * Ratio of the inner border size to
    	 * the lesser of the component width and height.
    	 */
    	public float borderFactor2 = 1/5f;
    	
    	/**
    	 * Ratio of the width of the outline (a line drawn in
    	 * the border) to the border width.
    	 */
    	public float outlineThicknessFactor = 1/5f; // fraction of the combined border
    	
    	/**
    	 * A flag indicating whether the
    	 * outline is to be drawn.
    	 */
    	public boolean showOutline = true;
    	
    	/**
    	 * A flag indicating whether the
    	 * image is to be drawn with antialiasing.
    	 */
    	public boolean antialias = true;
    	
    	/**
    	 * Returns a string representation of
    	 * this configuration object.
    	 * Used for debugging.
    	 */
    	public String toString() {
    		String base = getClass().getName();
			base = base.substring(base.lastIndexOf('.')+1);
			return base+"[" +
    			super.toString()+
    			",text="+text+
    			",font="+font+
    			"]";
    	}
		
		/**
    	 * Ratio of x coordinate, of the gradient starting point,
    	 * to the component width.
    	 */
    	public float point1x = 0.5f;
		
		/**
    	 * Ratio of y coordinate, of the gradient starting point,
    	 * to the component height.
    	 */
    	public float point1y = 0.25f;
		
		/**
    	 * Ratio of x coordinate, of the gradient ending point,
    	 * to the component width.
    	 */
    	public float point2x = 0.5f;
		
		/**
    	 * Ratio of y coordinate, of the gradient ending point,
    	 * to the component height.
    	 */
    	public float point2y = 0.7f;
    	
    	/**
    	 * Indicates which gradient point can currently be
    	 * changed using the user interface.
    	 * 0 = none, 1 = point1 (start point), 2 = point2 (end point).
    	 */
		public int editPoint = 1;	// 0 = none, 1 = point1, 2 = point2
    }

    // Required by ImageCreator
    public static Config createEmptyConfig() {
    	return new Config();
    }

    private Config cfg;
    private JPanel toolPanel;
    private JTextField textField;
    private ColorChooserBox colorBox1;
    private ColorChooserBox colorBox2;
    private ColorChooserBox bgColorBox;
    private ColorChooserBox outlineColorBox;
    private FontDialog fontDialog;
    private int sliderMax = 500;
    private float maxBorderFactor = 0.495f; // max sum of borderFactor's
    private JSlider borderSlider1;
    private JSlider borderSlider2;
    private JSlider outlineSlider;
    private JCheckBox outlineCheckBox;
    private JCheckBox antialiasCheckBox;
    private JCheckBox p1Button;
    private JCheckBox p2Button;

	private JComponent displayPanel;
	private JLayeredPane layers;
	private PointSelector pointSelector;

    /**
	 * Constructs a GradientText object.
	 */
    public GradientText() {
    	cfg = new Config();
    	init();
    }

    /**
	 * Constructs a GradientTextt object given its configuration.
	 * A constructor with this signature is required by the
	 * ImageCreator application.
	 */
    public GradientText(Config c) {
    	cfg = c;
    	init();
    }


    private void init() {
    	cfg.ensureMinSize(minSize);

		Dimension sz = cfg.getSize();
		//dpl("sz="+sz);
		setSize(sz);
    	setPreferredSize(sz);

		// Create the component to be displayed by the GUI
		//
		int[] x = getPoints();
		Point p = new Point(x[0], x[1]);
		if (cfg.editPoint == 2) {
			p.setLocation(x[2], x[3]);
		}
		pointSelector = new PointSelector(this, p.x, p.y);
		pointSelector.setEnabled(cfg.editPoint != 0);
		displayPanel = pointSelector;
		pointSelector.addPropertyChangeListener("point",
			new PropertyChangeListener() {
    			public void propertyChange(PropertyChangeEvent pce) {
    				Point newp = (Point) pce.getNewValue();
    				Point oldp = (Point) pce.getOldValue();
    				//dpl("newp = " + newp + ", oldp=" + oldp);
    				if (p1Button.isSelected()) {
    					setPoint1(newp);
    				}
    				else if (p2Button.isSelected()) {
    					setPoint2(newp);
    				}
    				else {
    					throw new Error(
    					"got point change " +
    					"with neither button selected, " +
    					"pointSelector.isEnabled()=" +
    					pointSelector.isEnabled()
    					);
    					
    				}		
    			}
    		});
		/*
		displayPanel = this;
		*/
    }

    private void setPoint1(Point p) {
    	//dpl("old point 1 cfg = " + cfg.point1x + ", " + cfg.point1y);
    	cfg.point1x = (float) p.x/getWidth();
        cfg.point1y = (float) p.y/getHeight();
        //dpl("new point 1 cfg = " + cfg.point1x + ", " + cfg.point1y);
    	fireStateChanged();
    }
    private void setPoint2(Point p) {
    	cfg.point2x = (float) p.x/getWidth();
        cfg.point2y = (float) p.y/getHeight();
        fireStateChanged();
    }

    private Point getPoint1() {
    	Point p = new Point();
    	p.setLocation(round(getWidth() *  cfg.point1x),
    	              round(getHeight() * cfg.point1y));
    	return p;
    }
    private Point getPoint2() {
    	Point p = new Point();
    	p.setLocation(round(getWidth() *  cfg.point2x),
    	              round(getHeight() * cfg.point2y));
    	return p;
    }

	private void getPoints(int[] x) {
		x[0] = round(getWidth() *  cfg.point1x);
		x[1] = round(getHeight() * cfg.point1y);
		x[2] = round(getWidth() *  cfg.point2x);
		x[3] = round(getHeight() * cfg.point2y);
	}
	private int[] getPoints() {
		int[] rv = new int[4];
		getPoints(rv);
		return rv;
	}

	private void addPointSelector() {
		layers.add(pointSelector, new Integer(1));	
	}

	private void removePointSelector() {
		layers.remove(pointSelector);	
	}

	/**
	 * Paints the image.  Overrides javax.swing.JComponent.
	 * Note that ImageModule.paintImage will cause this
	 * method to be invoked.
	 * @see ImageModule#paintImage
	 */
	public void paintComponent(Graphics g) {
	
		//dpl("preferred size = " + getPreferredSize());
		//dpl("size = " + getSize());
		//dpl(".");
		Graphics2D g2 = (Graphics2D) g;
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                             (cfg.antialias
                              ? RenderingHints.VALUE_ANTIALIAS_ON
                              : RenderingHints.VALUE_ANTIALIAS_OFF));
        g2.setRenderingHint(RenderingHints.KEY_RENDERING,
                             (cfg.antialias
                              ? RenderingHints.VALUE_RENDER_QUALITY
                              : RenderingHints.VALUE_RENDER_SPEED));

        int ww = getWidth();
        int hh = getHeight();
        //dpl("paintComponent: ww,hh = " + ww + "," + hh);
        g2.setBackground(cfg.bgColor);
        g2.clearRect(0,0,ww,hh);
        //Color origColor = g2.getColor();
        //
        //g2.setColor(Color.black);

        float borderSize1 = cfg.borderFactor1 * Math.min(ww,hh);
        float borderSize2 = cfg.borderFactor2 * Math.min(ww,hh);
        float borderSize = borderSize1 + borderSize2;
        if (cfg.showOutline) {
        	float thickness = cfg.outlineThicknessFactor * borderSize;
        	float inset = borderSize1;
        	if (borderSize1 < thickness/2) {
        		inset = thickness/2;
        	}
        	else if (borderSize2 < thickness/2) {
        		inset = borderSize - (thickness/2);
        	}
        	/*
			dpl("thickness="+thickness+
        	    ", borderSize1="+borderSize1+
        	    ", inset="+inset);
			*/
        	g2.setColor(cfg.outlineColor);
        	g2.setStroke(new BasicStroke( thickness ));
        	g2.draw(new Rectangle2D.Float(
        		inset, inset, ww-2*inset, hh-2*inset));
        }
        float lw = ww - 2*borderSize;
        float lh = hh - 2*borderSize;
        Shape s = getLabelShape(g2, cfg.text, cfg.font, lw, lh,
                                borderSize,
                                borderSize);
		g2.setPaint(
          new GradientPaint(getPoint1(), cfg.color1,
		                    getPoint2(), cfg.color2));
        g2.fill(s);

        //g2.setColor(origColor);
	}
	
	private float getBorderFactor(int which) {
		return which==0 ? cfg.borderFactor1 : cfg.borderFactor2;
	}
	private void setBorderFactor(int which, float val) {
		if (which==0)
			cfg.borderFactor1 = val;
		else
			cfg.borderFactor2 = val;
	}
	private void adjustBorderFactor(int which) {
		if (cfg.borderFactor1 + cfg.borderFactor2 > maxBorderFactor)
		{
			float f = maxBorderFactor - getBorderFactor((which+1)%2);
			setBorderFactor(which, f);
			JSlider sl = which==0 ? borderSlider1 : borderSlider2;
			int slval = round(sl.getMaximum() * f / maxBorderFactor);
   	    	sl.removeChangeListener(bsl);
   			sl.setValue(slval);
   			sl.addChangeListener(bsl);
   		}
   	}	
	private ChangeListener bsl = new ChangeListener() {
   		public void stateChanged(ChangeEvent e) {
   			JSlider src = (JSlider)e.getSource();
   			int which = src==borderSlider1 ? 0 : 1;
   			int other = (which + 1) % 2;
   			int v = src.getValue();
   			//dpl("slider " + which + " value = " + v);
   			//
   			float x = v * maxBorderFactor / src.getMaximum();
   			if (getBorderFactor(which) != x) {
   				setBorderFactor(which, x);
   				adjustBorderFactor(other);
   				fireStateChanged();
   			}
   		}
   	};


	/**
	 * Gets the component to be displayed by
	 * the user interface.  The returned component contains
	 * the module's image, and some of the "tools"
	 * for configuring the image.
	 * Overrides ImageModule.
	 * @see #getToolPanel
	 * @see ImageModule#getDisplayPanel
	 */
	public JComponent getDisplayPanel() {
	
		/*
		//////////////////////////////////////////////
		// unused code; for debug only
		if (false) {
			Dimension sz = cfg.getSize();
			displayPanel = new JPanel() {
				public void paintComponent(Graphics g) {
					dpl(" ---------- paint -------------");
				}
			};
			displayPanel.setPreferredSize(sz);
		}
		//////////////////////////////////////////////
        */

		//dpl("displayPanel is " + displayPanel.getClass());
		return displayPanel;
	}

	/**
	 * Paints the module's image.
	 * Overrides ImageModule.paintImage
	 * @see ImageModule#paintImage
	 */
	public void paintImage(Graphics g) {
		paintComponent(g);
	}

	/**
	 * Gets the tool panel, which provides a user interface
	 * for modifying the module.
	 * @returns the tool panel.
	 * @see #getDisplayPanel
	 * @see ImageModule#getToolPanel
	 */
	public JComponent getToolPanel() {
		if (toolPanel != null)
			return toolPanel;
		toolPanel = new JPanel();
		toolPanel.setLayout(new BorderLayout());
		Box vbox = Box.createVerticalBox();
		toolPanel.add(vbox, BorderLayout.CENTER);
		
		textField = new JTextField(cfg.text);
		textField.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				cfg.text = textField.getText();
				fontDialog.set(cfg.font, cfg.text); //
				fireStateChanged();	
			}
		});
		vbox.add(textField);
		
		fontDialog = new FontDialog(cfg.font, cfg.text);
		fontDialog.addPropertyChangeListener(new PropertyChangeListener() {
			public void propertyChange(PropertyChangeEvent e) {
				dpl("got property '" + e.getPropertyName() + "'");
				dpl("old value = '" + e.getOldValue() + "'");
				dpl("new value = '" + e.getNewValue() + "'");
				String p = e.getPropertyName();
				if (p.equals("font")) {
					cfg.font = new SimpleFont((Font)e.getNewValue());
					dpl("cfg.font = '" + cfg.font + "'");
					fireStateChanged();
				}
				else if (p.equals("text")) {
					cfg.text = (String)e.getNewValue();
					textField.setText(cfg.text);
					fireStateChanged();
				}
			}
		});
		JButton fontButton = new JButton("font");
		fontButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				//if (!fontDialog.isShowing())
				fontDialog.show();
			}
		});
		vbox.add(fontButton);
		
		colorBox1 = new ColorChooserBox(
			"color 1: ", BoxLayout.X_AXIS, cfg.color1);
		colorBox1.addPropertyChangeListener(
			new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent e) {
					try {
						Color c = (Color)e.getNewValue();
						if (!c.equals(cfg.color1)) {
							//dpl("new color = " + c);
							cfg.color1 = new SimpleColor(c);
							fireStateChanged();
						}
					}
					catch (ClassCastException ce) {}
				}
		});
		vbox.add(colorBox1);
		
		colorBox2 = new ColorChooserBox(
			"color 2: ", BoxLayout.X_AXIS, cfg.color2);
		colorBox2.addPropertyChangeListener(
			new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent e) {
					try {
						Color c = (Color)e.getNewValue();
						if (!c.equals(cfg.color2)) {
							cfg.color2 = new SimpleColor(c);
							fireStateChanged();
						}
					}
					catch (ClassCastException ce) {}
				}
		});
		vbox.add(colorBox2);
		
		bgColorBox = new ColorChooserBox(
			"background: ", BoxLayout.X_AXIS, cfg.bgColor);
		bgColorBox.addPropertyChangeListener(
			new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent e) {
					try {
						Color c = (Color)e.getNewValue();
						if (!c.equals(cfg.bgColor)) {
							cfg.bgColor = new SimpleColor(c);
							fireStateChanged();
						}
					}
					catch (ClassCastException ce) {}
				}
		});
		vbox.add(bgColorBox);
		
		outlineColorBox = new ColorChooserBox(
			"outline: ", BoxLayout.X_AXIS, cfg.outlineColor);
		outlineColorBox.addPropertyChangeListener(
			new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent e) {
					try {
						Color c = (Color)e.getNewValue();
						if (!c.equals(cfg.outlineColor)) {
							cfg.outlineColor = new SimpleColor(c);
							fireStateChanged();
						}
					}
					catch (ClassCastException ce) {}
				}
		});
		vbox.add(outlineColorBox);
		
		borderSlider1 = new JSlider(0,sliderMax,
			round(sliderMax*cfg.borderFactor1/maxBorderFactor));
		borderSlider1.addChangeListener(bsl);
		vbox.add(new JLabel("Outer border:"));
		vbox.add(borderSlider1);

		borderSlider2 = new JSlider(0,sliderMax,
			round(sliderMax*cfg.borderFactor2/maxBorderFactor));
		//dpl("slider 2 initial value = " + borderSlider2.getValue());
		borderSlider2.addChangeListener(bsl);
		vbox.add(new JLabel("Inner border:"));
		vbox.add(borderSlider2);

		
		outlineSlider = new JSlider(
			0,sliderMax,round(sliderMax*cfg.outlineThicknessFactor));
		outlineSlider.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				int v = outlineSlider.getValue();
				float max = 1.0f;
				float x = v * max / outlineSlider.getMaximum();
				cfg.outlineThicknessFactor = x;
				fireStateChanged();
			}
		});
		vbox.add(new JLabel("Outline thickness:"));
		vbox.add(outlineSlider);
		
		outlineCheckBox =
			new JCheckBox("Show outline", cfg.showOutline);
		outlineCheckBox.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				boolean v = outlineCheckBox.isSelected();
				if (v != cfg.showOutline) {
					cfg.showOutline = v;
					fireStateChanged();
				}
			}
		});
		
		if (true) {
		//outlineCheckBox.setHorizontalAlignment(SwingConstants.CENTER);
		outlineCheckBox.setAlignmentX(0.5f);
		vbox.add(outlineCheckBox);
		} else {
		Box row = Box.createHorizontalBox();
		row.add(Box.createHorizontalGlue());
		row.add(outlineCheckBox);
		row.add(Box.createHorizontalGlue());
		vbox.add(row);
		}
		
		antialiasCheckBox =
			new JCheckBox("antialias", cfg.antialias);
		antialiasCheckBox.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				boolean v = antialiasCheckBox.isSelected();
				if (v != cfg.antialias) {
					cfg.antialias = v;
					fireStateChanged();
				}
			}
		});
		antialiasCheckBox.setAlignmentX(0.5f);
		vbox.add(antialiasCheckBox);
		
		p1Button = new JCheckBox("change point 1");
		p2Button = new JCheckBox("change point 2");
		p1Button.setSelected(cfg.editPoint == 1);
		p2Button.setSelected(cfg.editPoint == 2);
		p1Button.addActionListener(pointButtonListener);
		p2Button.addActionListener(pointButtonListener);
		p1Button.setAlignmentX(0.5f);
		p2Button.setAlignmentX(0.5f);
		vbox.add(p1Button);
		vbox.add(p2Button);
			
		return toolPanel;
		
	}
	
	private ActionListener pointButtonListener = new ActionListener() {
	
		public void actionPerformed(ActionEvent e) {
		
			PointSelector ps = pointSelector;
			//dpl("ps="+ps);
			JCheckBox b = (JCheckBox)e.getSource();
			
			// Deselect the other button if this one is selected
			//
			JCheckBox other = (b==p1Button ? p2Button : p1Button);
			if (b.isSelected()) {
				other.removeActionListener(this);
				other.setSelected(false);
				other.addActionListener(this);
			}
			
			if (p1Button.isSelected()) {
				cfg.editPoint = 1;
				if (ps != null) {
					ps.setPoint(getPoint1());
					ps.setEnabled(true);
				}
			}
			else if (p2Button.isSelected()) {
				cfg.editPoint = 2;
				if (ps != null){
					//dpl("setting ps point to " + getPoint2());
					ps.setPoint(getPoint2());
					ps.setEnabled(true);
				}
			}
			else {
				cfg.editPoint = 0;
				if (ps != null)
					ps.setEnabled(false);
			}
			
			
			
		}
	};

	/**
	 * Gets the configuration object.
	 */
	public ModuleConfig getConfig() {
		//dpl("getConfig: color1 = " + cfg.color1);
		return cfg;
	}

	/**
	 * Sets the configuration object.
	 *
	 * To do: This method is not currently [13 Jan 2002] used.
	 * Get rid of it?
	 * If it were used, it contains a bug: it should cause the
	 * displayPanel to be resized.
	 */
	public void setConfig(ModuleConfig mc) {
		
		if (mc instanceof Config) {
			cfg = (Config) mc;
			
			//setPreferredSize(cfg.getSize());
			setSize(cfg.getSize());
			//dpl("setConfig: preferredSize=" + getPreferredSize());
			dpl("setConfig: size=" + getSize());
		}
		else
			perror("setConfig: Wrong type: " + mc.getClass().getName());
	}
	
	/**
	 * Computes a shape from a string of text.
	 * Note: This is a utility which might best be moved to a library.
	 *
	 * @param	g2	graphics context
	 * @param	lstring	the string for which a shape is to be created.
	 * @param	font	the font
	 * @param	labelWidth	the width of the bounding rectangle
	 *           for the shape.
	 * @param	labelHeight	the height of the bounding rectangle
	 *           for the shape.
	 * @param	xOffset	the distance from the origin to the left edge
	 *           of the bounding rectangle.
	 * @param	yOffset	the distance from the origin to the top edge
	 *           of the bounding rectangle.
	 *
	 */
    private static Shape getLabelShape(Graphics2D g2,
                               String lstring,
                               Font font,
                               double labelWidth,
                               double labelHeight,
                               double xOffset,
                               double yOffset)
    {
    	err.println("getLabelShape: w=" + labelWidth + ", h=" + labelHeight);
    	if (lstring == null || labelWidth == 0 || labelHeight == 0) {
    		/*
			err.println("getLabelShape: lstring=" + lstring
    		    + ", w=" + labelWidth + ", h=" + labelHeight);
			*/
    		return null;
    	}
	
    	GlyphVector v;
    	v = font.createGlyphVector(g2.getFontRenderContext(), lstring);
    	Shape rv = v.getOutline();
    	Rectangle2D b = rv.getBounds2D();
    	double xscale = labelWidth/b.getWidth();
    	double yscale = labelHeight/b.getHeight();
    	
    	AffineTransform a = AffineTransform.getTranslateInstance(xOffset, yOffset);
    	a.scale(xscale, yscale);
    	// translate origin to upper left
    	a.translate(-b.getMinX(), -b.getMinY());
    	
    	rv = a.createTransformedShape(rv);
    	err.println("getLabelShape: bounds=" + rv.getBounds2D() );
    	return rv;
    }
}
