/*
Image Creator: Creates configurable images
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package scotthaywood.imageCreator.modules;

import java.util.*;
import java.awt.*;
import java.awt.font.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import scotthaywood.ui.*;
import scotthaywood.imageCreator.*;

/**
 * MagnifiedText is a type of module for the ImageCreator
 * application.
 * Its image consists of lines of text,
 * with an optional "magnifying glass" over them.
 *
 * @author Scott Haywood
 */
public class MagnifiedText extends ImageModule {

	private void dpl(String s) {
		System.err.println(getClass().getName() + ": " + s);
	}

	private int round(double d) {
   		return (int)Math.round(d);
   	}
	
	private static Dimension defaultSize = new Dimension(200,200);
	private static Dimension minSize = new Dimension(10,10);
	
	/**
	 * Creates an empty configuration object.
	 * This method is required by ImageCreator.
	 * @returns the empty configuration object.
	 */
	public static Config createEmptyConfig() {
		return new Config();
	}
	
	/**
	 * The inner class <code>Config</code>
	 * is the type of configuration object for this module.
	 * Its public fields contain all the information needed
	 * to save and restore a module.
	 */
    public static class Config extends ImageModuleConfig {

		private static SimpleFont defaultFont =
			new SimpleFont("Lucida Sans", Font.PLAIN, 16);

		private static String defaultText = "This is a test\n" +
				"1111111111111111\n" + 
				"2222222222222222\n" +
				"3333333333333333\n" +
				"4444444444444444\n" +
				"5555555555555555\n" +
				"6666666666666666\n" +
				"7777777777777777\n" +
				"8888888888888888\n" +
				"9999999999999999\n";

		private static Magnifier defaultMagnifier = new Magnifier();

		/**
		 * The font in which the text is displayed
		 */
		public SimpleFont font = defaultFont;
		
		/**
		 * The magnifying glass.
		 */
		public Magnifier magnifier = defaultMagnifier;
		
		/**
		 * An indicator as to whether the magnifying glass is displayed.
		 */
		public boolean show_magnifier = true;
		
		/**
		 * An indicator as to whether the font
		 * size is changed when the module width is changed.
		 */
		public boolean resize_font = true;
		
		/**
		 * The text to be displayed.
		 */
		public String text = defaultText;
		
		/**
		 * Constructs a configuration object.
		 */
		public Config() {
			super.setSize(defaultSize);
		}
    }

    private Config cfg;
    private JPanel toolPanel;
    private JRadioButton editButton;
	private JRadioButton magButton;
	private JCheckBox resizeFontButton;

    private JLayeredPane layers = new JLayeredPane();
    private JTextPane textPane;
    private Magnifier pip; // the name pip is from "picture in picture"
	private BasicStroke circleStroke;
	private int oldTextPaneWidth;

	
	//
	// sets the textPane, given its size, font, and text.
	//
	private void setTextPane(Dimension d, Font f, String s) {
		JTextPane t = new JTextPane();
    	t.setSize(d);
		oldTextPaneWidth = d.width;
		t.setFont(f);
		t.setText(s);
		t.setLocation(0,0);
		
		// when the JTextPane document changes, copy it
		// to the configuration object, and fire an event.
		//
		Document doc = t.getDocument();
		doc.addDocumentListener(new DocumentListener() {
			private void update() {
				cfg.text = textPane.getText();
				fireStateChanged();
			}
			public void changedUpdate(DocumentEvent e) {
				//dpl("changed update");
				update();
			}
			public void insertUpdate(DocumentEvent e) {
				//dpl("insert update");
				update();
			}
			public void removeUpdate(DocumentEvent e) {
				//dpl("remove update");
				update();
			}
		});
		if (layers == null) throw new Error("layers not set");
		if (textPane != null) {
			layers.remove(textPane);
		}
		layers.add(t, new Integer(0));
		textPane = t;
	}

	private Map fontAttr = new HashMap(4);
	{
	//fontAttr.put(TextAttribute.WEIGHT, TextAttribute.WEIGHT_REGULAR);
	//fontAttr.put(TextAttribute.POSTURE, TextAttribute.POSTURE_REGULAR);
	fontAttr.put(TextAttribute.SIZE, new Float(18f));
	fontAttr.put(TextAttribute.FAMILY, "Lucida Sans");
	}

	//
	// changes the size of the text pane.
	//
	private void changeTextPaneSize(Dimension d) {

		// Change the size of the JLayeredPane
		//
		layers.setSize(d);

		if (cfg.resize_font) {
    		// Change the size of the JTextPane, and
    		// change its font size, proportional to the change
    		// in width.
    		//
    		Font f = textPane.getFont();
    		float oldSize = f.getSize2D();
    		float newSize = oldSize * d.width / oldTextPaneWidth;
    		fontAttr.put(TextAttribute.SIZE, new Float(newSize));
    		Font newf = new Font(fontAttr);
    		
    		// Save new font in config.
    		// Note that fireStateChanged() is not needed by
    		// the ImageCreator application, because it is
    		// called whenever an ImageModule is resized.
    		// (Another fireStateChanged() might be needed by
    		// some other hypothetical application.)
    		cfg.font = new SimpleFont(newf);
    		textPane.setFont(newf);
		}
		
		textPane.setSize(d);
		oldTextPaneWidth = d.width; // remember this width
		
		// Move and resize the magnifier
		//
		doPipLayout(d);
	}

	private void doPipLayout(Dimension d) {
		pip.setLocation(0,0);
		pip.setSize(d);
	}

	private void addPip() {
		if (layers == null) throw new Error("layers not set");
    	layers.add(pip, new Integer(1));
	}
	private void removePip() {
		if (layers == null) throw new Error("layers not set");
		layers.remove(pip);
	}


	/**
	 * Constructs a MagnifiedText object given its configuration.
	 * A constructor with this signature is required by the
	 * ImageCreator application.
	 */
    public MagnifiedText(Config config) {
		cfg = config;
		init();
	}

    /**
	 * Constructs a MagnifiedText object.
	 */
    public MagnifiedText() {
		cfg = new Config();
		init();
	}

	private void init() {
		
		 
		cfg.ensureMinSize(minSize);
		
		Dimension d = cfg.getSize();
		setPreferredSize(d);
    	
    	setLayout(null);
    	
    	layers.setSize(d);
    	layers.setLocation(0,0);
    	add(layers);
    	
    	setTextPane(d, cfg.font, cfg.text);
		cfg.magnifier.setView(textPane);

		// When the size of this component changes,
		// adjust the size of the children
		//
		ComponentAdapter resizeListener = new ComponentAdapter() {
			public void componentResized(ComponentEvent e) {
				changeTextPaneSize(getSize());
			}
		};
		addComponentListener(resizeListener);
    	
    	pip = cfg.magnifier;
		pip.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				fireStateChanged();
			}
		});
		pip.setView(textPane);
		doPipLayout(d);
		if (cfg.show_magnifier) {
			addPip();
		}
    }


	/**
	 * Get the tool panel, which provides a user interface
	 * for modifying the module.
	 * @returns the tool panel.
	 * @see ImageModule#getToolPanel
	 */
	public JComponent getToolPanel() {
	
		// return null;  
		
		if (toolPanel != null)
			return toolPanel;
		toolPanel = new JPanel();
		toolPanel.setLayout(new BorderLayout());
		Box box = Box.createVerticalBox();
		toolPanel.add(box, BorderLayout.CENTER); 

				
		// The text pane is editable only when the
		// magnifier is not shown
		//
		textPane.setEnabled(!cfg.show_magnifier);
		ButtonGroup bgroup = new ButtonGroup();
		editButton = new JRadioButton("edit", !cfg.show_magnifier);
		magButton = new JRadioButton("magnifier", cfg.show_magnifier);
		Box bbox = Box.createHorizontalBox();
		bbox.add(editButton);
		bbox.add(magButton);
		bgroup.add(editButton);
		bgroup.add(magButton);
		box.add(bbox);
		ActionListener bListener = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				boolean v = magButton.isSelected();
				//dpl("mag enabled = " + v);
				if (v != cfg.show_magnifier) {
					cfg.show_magnifier = v;
					textPane.setEnabled(!v);
					if (v) {
						addPip();
					}
					else {
						removePip();
					}
					fireStateChanged();
				}
			}
		};
		magButton.addActionListener(bListener);
		editButton.addActionListener(bListener);
		
		resizeFontButton = new JCheckBox("font grows w/ width",
		                                 cfg.resize_font);
		resizeFontButton.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				boolean v = resizeFontButton.isSelected();
				if (v != cfg.resize_font) {
					cfg.resize_font = v;
					fireStateChanged();
				}
			}
		});
		
		box.add(pip.getToolPanel());
		
		//resizeFontButton.setAlignmentX(0f);
		Box lbox = Box.createHorizontalBox();
		lbox.add(resizeFontButton);
		lbox.add(Box.createHorizontalGlue());
		box.add(lbox);

		return toolPanel;
	}
	
	/**
	 * Gets the configuration object.
	 * @returns the configuration object
	 */
	public ModuleConfig getConfig() {
		return cfg;
	}
	
	/**
	 * Sets the configuration object.
	 * @param	mc	the new configuration object.
	 * @see ImageModule#setConfig
	 * @see #getConfig
	 */
	public void setConfig(ModuleConfig mc) {
		if (mc instanceof Config) {
			cfg = (Config) mc;
			init();
		}
		else
			perror("setConfig: Wrong type: " + mc.getClass().getName());
	}
	
}
