/*
Image Creator: Creates configurable images
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package scotthaywood.imageCreator.modules;

import java.awt.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import javax.swing.*;
import scotthaywood.ui.*;
import scotthaywood.imageCreator.*;

/**
 * SingleColor is a type of module for the ImageCreator
 * application.
 * Its image consists of a rectangle of one color.
 * Its user interface provides a way of changing the color.
 *
 * @author Scott Haywood
 */
public class SingleColor extends ImageModule {

	private static Dimension defaultSize = new Dimension(200,200);
	private static Dimension minSize = new Dimension(1,1);
	
    /**
	 * The inner class <code>Config</code>
	 * is the type of configuration object for this module.
	 * Its public fields contain all the information needed
	 * to save and restore a module.
	 */
    public static class Config extends ImageModuleConfig {

    	private static SimpleColor defaultColor =
    		 new SimpleColor(0xFF0000);
    		
    	public SimpleColor color = defaultColor;
    	
    	public Config() {
    		super.setSize(defaultSize);
    	}
    }

    private Config cfg;
    private JPanel toolPanel;
    private ColorChooserBox colorBox;

    /**
	 * Constructs a SingleColor object.
	 * A default constructor is required by ImageCreator.
	 */
    public SingleColor() {
    	cfg = new Config();
    	init();
    }

    /**
	 * Constructs a SingleColor object given its configuration.
	 * A constructor with this signature is required by ImageCreator.
	 */
    public SingleColor(Config c) {
    	cfg = c;
    	init();
    }

    /**
	 * Creates an empty configuration object.
	 * This method is required by ImageCreator.
	 * @returns the empty configuration object.
	 */
	public static Config createEmptyConfig() {
		return new Config();
	}
	
    private void init() {
    	cfg.ensureMinSize(minSize);
    	setPreferredSize(cfg.getSize());
    }	

    private void setColor(Color c) {
    	if (c.getRGB() != cfg.color.getRGB()) {
    		cfg.color = new SimpleColor(c);
    		//repaint(); // superclass handles repaint
    		fireStateChanged();
    	}
    }

	/**
	 * Paints the image.  Overrides javax.swing.JComponent.
	 * Note that ImageModule.paintImage will cause this
	 * method to be invoked.
	 * @see ImageModule#paintImage
	 */
	public void paintComponent(Graphics graphics) {
		/*
		perror(getClass().getName() + ".paintComponent:" +
			" size=" + getSize());
		*/
		Graphics2D g = (Graphics2D) graphics;
		g.setBackground(cfg.color);
		g.clearRect(0,0,getWidth(), getHeight());
	}
	
	/**
	 * Get the tool panel, which provides a user interface
	 * for modifying the module.
	 * @returns the tool panel.
	 * @see ImageModule#getToolPanel
	 */
	public JComponent getToolPanel() {
		if (toolPanel != null)
			return toolPanel;
		toolPanel = new JPanel();
		toolPanel.setLayout(new BorderLayout());
    	colorBox = new ColorChooserBox(
    	                               "color:",
    	                               BoxLayout.X_AXIS,
    	                               cfg.color);
    	toolPanel.add(colorBox);
    	colorBox.addPropertyChangeListener("color",
    		new PropertyChangeListener() {
    			public void propertyChange(PropertyChangeEvent pce) {
    				Color newColor = (Color) pce.getNewValue();
    				setColor(newColor);		
    			}
    		});
		return toolPanel;
	}
	
	/**
	 * Gets the configuration object.
	 * @returns the configuration object
	 */
	public ModuleConfig getConfig() {
		return cfg;
	}
	
	/**
	 * Sets the configuration object.
	 * @param	mc	the new configuration object.
	 * @see ImageModule#setConfig
	 * @see #getConfig
	 */
	public void setConfig(ModuleConfig mc) {
		if (mc instanceof Config) {
			cfg = (Config) mc;
			setPreferredSize(cfg.getSize());
		}
		else
			perror("setConfig: Wrong type: " + mc.getClass().getName());
	}
	
}
