/*
Image Creator: Creates configurable images
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package scotthaywood.imageCreator.modules;

import java.awt.*;
import java.awt.event.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import javax.swing.*;
import javax.swing.border.*;
import scotthaywood.ui.*;
import scotthaywood.imageCreator.*;

/**
 * SingleColorWithBorder is a type of module for the ImageCreator
 * application.
 * Its image consists of a rectangle of one color,
 * with a titled border.
 * Its user interface provides a way of changing the color
 * and the title.
 *
 * @author Scott Haywood
 */
public class SingleColorWithBorder extends ImageModule {

	private static Dimension defaultSize = new Dimension(200,100);
	private static Dimension minSize = new Dimension(1,1);
	
    /**
	 * The inner class <code>Config</code>
	 * is the type of configuration object for this module.
	 * Its public fields contain all the information needed
	 * to save and restore a module.
	 */
    public static class Config extends ImageModuleConfig {
    	static SimpleColor dcolor = new SimpleColor(0xFF0000);

    	public String title = "";
    	public SimpleColor color = dcolor;

		public Config() {
			super.setSize(defaultSize);
		}
    }

	/**
	 * Creates an empty configuration object.
	 * This method is required by ImageCreator.
	 * @returns the empty configuration object.
	 */
	public static Config createEmptyConfig() {
    	return new Config();
    }

    private Config cfg;
	private JPanel toolPanel;
    private ColorChooserBox colorBox;
    private JTextField titleField;
    private TitledBorder border;

    /**
	 * Constructs a SingleColorWithBorder object.
	 * A default constructor is required by ImageCreator.
	 */
    public SingleColorWithBorder() {
    	
		cfg = new Config();
		init();
	}
	
	/**
	 * Constructs a SingleColorWithBorder object given its configuration.
	 * A constructor with this signature is required by ImageCreator.
	 */
    public SingleColorWithBorder(Config c) {
		cfg = c;
		init();
	}

	private void init() {
		
		cfg.ensureMinSize(minSize);
		setPreferredSize(cfg.getSize());
    	
    	// create the border
		if (cfg.title == null || cfg.title.length()==0) {
        	String t = getClass().getName();
        	cfg.title = t.substring(t.lastIndexOf('.')+1);
        }
    	border = BorderFactory.createTitledBorder(cfg.title);
        setBorder(border);
    }

	

    private void setColor(Color c) {
    	if (c.getRGB() != cfg.color.getRGB()) {
    		cfg.color = new SimpleColor(c);
    		//repaint(); // superclass handles repaint
    		fireStateChanged();
    	}
    }

	/**
	 * Paints the image.  Overrides javax.swing.JComponent.
	 * Note that ImageModule.paintImage will cause this
	 * method to be invoked.
	 * @see ImageModule#paintImage
	 */
	public void paintComponent(Graphics graphics) {
		/*
		perror(getClass().getName() + ".paintComponent:" +
			" size=" + getSize());
		*/
		Graphics2D g = (Graphics2D) graphics;
		g.setBackground(cfg.color);
		g.clearRect(0,0,getWidth(), getHeight());
	}
	
	/**
	 * Get the tool panel, which provides a user interface
	 * for modifying the module.
	 * @returns the tool panel.
	 * @see ImageModule#getToolPanel
	 */
	public JComponent getToolPanel() {
		if (toolPanel != null)
			return toolPanel;
		toolPanel = new JPanel();
		Box box = Box.createVerticalBox();
		toolPanel.add(box);
	
		Box lbox = Box.createHorizontalBox();
        lbox.add(new JLabel("title:", SwingConstants.LEFT));
        lbox.add(Box.createHorizontalGlue());
		box.add(lbox);
				
		titleField = new JTextField(border.getTitle());
		box.add(titleField);
		titleField.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				String s = titleField.getText();
				if (!cfg.title.equals(s)) {
					// update the Config object and the border
					border.setTitle(cfg.title = s);
					fireStateChanged();
				}	
			}
		});
		
    	colorBox = new ColorChooserBox(
    	                               "color:",
    	                               BoxLayout.X_AXIS,
    	                               cfg.color);
    	box.add(colorBox);
    	colorBox.addPropertyChangeListener("color",
    		new PropertyChangeListener() {
    			public void propertyChange(PropertyChangeEvent pce) {
    				Color newColor = (Color) pce.getNewValue();
    				setColor(newColor);		
    			}
    		});
		return toolPanel;
	}
	
	/**
	 * Gets the configuration object.
	 * @returns the configuration object
	 */
	public ModuleConfig getConfig() {
		return cfg;
	}
	
	/**
	 * Sets the configuration object.
	 * @param	mc	the new configuration object.
	 * @see ImageModule#setConfig
	 * @see #getConfig
	 */
	public void setConfig(ModuleConfig mc) {
		if (mc instanceof Config) {
			cfg = (Config) mc;
			setPreferredSize(cfg.getSize());
		}
		else
			perror("setConfig: Wrong type: " + mc.getClass().getName());
	}
	
}
