/*
Image Creator: Creates configurable images
Copyright (C) 2005, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package scotthaywood.imageCreator.modules;

import java.util.HashMap;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.geom.Ellipse2D;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import javax.swing.*;
import scotthaywood.ui.*;
import scotthaywood.asciiIO.*;
import scotthaywood.imageCreator.*;

/**
 * TwoEllipse is a type of module for the ImageCreator
 * application.
 * Its image consists of a background rectangle,
 * with a centered ellipse,
 * and a second ellipse of inside the first one.
 * Its user interface provides a way of changing the colors
 * and sizes of the ellipses.
 *
 * @author Scott Haywood
 */
public class TwoEllipse extends ImageModule {

	private static Dimension defaultSize = new Dimension(200,200);
	private static Dimension minSize = new Dimension(1,1);
	
    /**
	 * The inner class <code>EllipseData</code>
	 * is for holding descriptions of the ellipses.
	 */
	public static class EllipseData implements AsciiIO.Externalizable
	{
		public float margin; // pixels around this ellipse
		public SimpleColor color;
		public EllipseData() {
			this(0f, SimpleColor.white);
		}
		public EllipseData(float m, SimpleColor c) {
			margin = m;
			color = c;
		}
		public Color getColor() { return color; }
		public void setColor(Color c) { color = new SimpleColor(c); }
		public float getMargin() { return margin; }
		public void setMargin(float f) { margin = f; }
	};

    /**
	 * The inner class <code>Config</code>
	 * is the type of configuration object for this module.
	 * Its public fields contain all the information needed
	 * to save and restore a module.
	 */
    public static class Config extends ImageModuleConfig {

    	private static SimpleColor defaultBG = new SimpleColor(0xFF0000);
    	private static SimpleColor colorZero = new SimpleColor(0x000000);
    	private static SimpleColor colorOne = new SimpleColor(0xFF0000);
		private static final float marginZero = 5.0f;
		private static final float marginOne = 3.0f;
    		
    	public SimpleColor bgColor = defaultBG;

		public EllipseData[] edata = {
			new EllipseData(marginZero, colorZero),
			new EllipseData(marginOne, colorOne)
		};
    	
    	public Config() {
    		super.setSize(defaultSize);
    	}
    }

    private Config cfg;
    private JPanel toolPanel;
    private ColorChooserBox bgColorBox;
    private ColorChooserBox[] ellipseColorBox;

    /**
	 * Constructs a TwoEllipse object.
	 * A default constructor is required by ImageCreator.
	 */
    public TwoEllipse() {
    	cfg = new Config();
    	init();
    }

    /**
	 * Constructs a TwoEllipse object given its configuration.
	 * A constructor with this signature is required by ImageCreator.
	 */
    public TwoEllipse(Config c) {
    	cfg = c;
    	init();
    }

    /**
	 * Creates an empty configuration object.
	 * This method is required by ImageCreator.
	 * @returns the empty configuration object.
	 */
	public static Config createEmptyConfig() {
		return new Config();
	}
	
    private void init() {
    	cfg.ensureMinSize(minSize);
    	setPreferredSize(cfg.getSize());
    }	

    private void setColor(Color c) {
    	if (c.getRGB() != cfg.bgColor.getRGB()) {
    		cfg.bgColor = new SimpleColor(c);
    		//repaint(); // superclass handles repaint
    		fireStateChanged();
    	}
    }

    private void setColor(EllipseData e, Color c) {
		
    	if (c.getRGB() != e.getColor().getRGB()) {
    		e.setColor(c);
    		//repaint(); // superclass handles repaint
    		fireStateChanged();
    	}
    }

	/**
	 * Paints the image.  Overrides javax.swing.JComponent.
	 * Note that ImageModule.paintImage will cause this
	 * method to be invoked.
	 * @see ImageModule#paintImage
	 */
	public void paintComponent(Graphics graphics) {
		/*
		perror(getClass().getName() + ".paintComponent:" +
			" size=" + getSize());
		*/
		Graphics2D g = (Graphics2D) graphics;

   		final Dimension dim = new Dimension();
   		getSize(dim);
   		int width = dim.width;
   		int height = dim.height;
   		float x = width/2;
   		float y = height/2;

		g.setBackground(cfg.bgColor);
		g.clearRect(0,0,getWidth(), getHeight());

		float ew = width;
		float eh = height;
		for (int i = 0; i< cfg.edata.length; i++) { 

			EllipseData edat = cfg.edata[i];
			
			ew -= edat.margin * 2; if (ew < 0) ew = 0;
			eh -= edat.margin * 2; if (eh < 0) eh = 0;
			g.setColor(edat.color);
			g.fill(new Ellipse2D.Float((width-ew)/2, (height-eh)/2, ew, eh));
			
		}
	}

   	private static int round(double d) {
   		return (int)Math.round(d);
   	}
	
	/**
	 * Get the tool panel, which provides a user interface
	 * for modifying the module.
	 * @returns the tool panel.
	 * @see ImageModule#getToolPanel
	 */
	public JComponent getToolPanel() {


		if (toolPanel != null)
			return toolPanel;
		toolPanel = new JPanel();
		toolPanel.setLayout(new BorderLayout());
		Box vbox = Box.createVerticalBox();
		toolPanel.add(vbox, BorderLayout.CENTER);
    	bgColorBox = new ColorChooserBox(
    	                               "background color:",
    	                               BoxLayout.X_AXIS,
    	                               cfg.bgColor);
    	vbox.add(bgColorBox);
    	bgColorBox.addPropertyChangeListener("color",
			new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent pce) {
    				Color newColor = (Color) pce.getNewValue();
    				setColor(newColor);		
    			}
    		});
		
		for (int i=0; i<cfg.edata.length; i++) {
			EllipseData edat = cfg.edata[i];
			ColorChooserBox ccb = new ColorChooserBox(
									   "ellipse " + i + " color:",
									   BoxLayout.X_AXIS,
									   edat.color
									   );
			ccb.setTag(edat);
			ccb.addPropertyChangeListener( new PropertyChangeListener() {
				public void propertyChange(PropertyChangeEvent pce) {
					ColorChooserBox b = (ColorChooserBox) pce.getSource();
					EllipseData e;
					e = (EllipseData) b.getTag();
					Color newColor = (Color) pce.getNewValue();
					setColor(e, newColor);
				}
			});
			
			Box hbox = Box.createHorizontalBox();
			hbox.add(ccb);
			hbox.add(Box.createHorizontalStrut(5));
			hbox.add(new JLabel("margin:"));
			JTextField tf = new JTextField(
				String.valueOf(edat.getMargin()),5);
			origBackground = tf.getBackground();
			hbox.add(tf);
			vbox.add(hbox);
			ActionListener alis = new ActionListener() {
				public void actionPerformed(ActionEvent ae) {
					EllipseData ell;
					ell = (EllipseData) toEllipseData.get(this);
					JTextField f = (JTextField) ae.getSource();
					float val = 0f;
					f.setBackground(origBackground);
					try {
						val = Float.parseFloat(f.getText());
					}
					catch (NumberFormatException nf) {
						val = ell.getMargin();
						f.setBackground(Color.pink);
					}
					if (val != ell.getMargin()) {
						// update the Config object and repaint
						ell.setMargin(val);
						fireStateChanged();
					}	
				}
			};
			tf.addActionListener(alis);
			toEllipseData.put(alis, edat);

		}
		return toolPanel;
	}
	private Color origBackground;
	private HashMap toEllipseData = new HashMap(5);
	
	/**
	 * Gets the configuration object.
	 * @returns the configuration object
	 */
	public ModuleConfig getConfig() {
		return cfg;
	}
	
	/**
	 * Sets the configuration object.
	 * @param	mc	the new configuration object.
	 * @see ImageModule#setConfig
	 * @see #getConfig
	 */
	public void setConfig(ModuleConfig mc) {
		if (mc instanceof Config) {
			cfg = (Config) mc;
			setPreferredSize(cfg.getSize());
		}
		else
			perror("setConfig: Wrong type: " + mc.getClass().getName());
	}
	
}
