/*
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
package scotthaywood.ui;

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import scotthaywood.ui.ColorIcon;

/**
 * A <code>ColorChooserBox</code> provides an interface
 * to <code>javax.swing.JColorChooser</code>.
 * It displays a text label,
 * a small square of color (the "swatch"), and a button.
 * When either the swatch or the button is clicked, it
 * invokes a JColorChooser dialog to allow the color
 * to be changed.  It fires a property change event if the
 * color is changed.
 *
 * @author Scott Haywood
 * @version $Revision: 1.6 $
 */
public class ColorChooserBox extends Box implements HasTag {
 	
	/**
	 * Constructs a color chooser box, with no text label,
	 * with horizontal axis, and the default color.
	 */
	public ColorChooserBox() {
		this(null, BoxLayout.X_AXIS, defaultColor);
	}
	
	/**
	 * Constructs a color chooser box, using the default property name.
	 */
	public ColorChooserBox(String title, int axis, Color initialColor) {
		this(title, axis, initialColor, defaultPropertyName);
	}

	/**
	 * Constructs a color chooser box.
	 *
	 * @param	title	the string to be displayed in a text label.
	 * @param	axis	the orientation of the box.
	 * @param	initialColor	the initial color of the swatch.
	 */
	public ColorChooserBox(
		String title, int axis, Color initialColor, String propertyName) {
	
		super(axis);
		titleLabel.setText(title != null ? title : "");
		color = (initialColor != null) ?  initialColor : defaultColor;
		colorIcon.setColor(color);
		colorButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				chooseColor();
			}
		});
		swatch.addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent e) {
				chooseColor();
			}
		});
		add(titleLabel);
		add(swatch);
		add(colorButton);

		myPropertyName = propertyName;
	}

	private void chooseColor() {
			
		Color c = JColorChooser.showDialog(ColorChooserBox.this,
				                              "Choose Color", color);
		if (c != null) {
			Color oldColor = color;
			color = c;
			setSwatchColor(color);
			firePropertyChange(myPropertyName, oldColor, color);
		}
	}
	
	/**
	 * Returns the current color.
	 */
	public Color getColor() { return color; }
	
	/**
	 * Sets the current color without generating an event.
	 */
	public void setColorNoEvent(Color c) {
		color = c;
		setSwatchColor(color);
	}

	
    // ------------------------------------------
	private Object tag;
	public void setTag(Object t) { tag = t; }
	public Object getTag() { return tag; }

    // ------------------------------------------

    public static Color defaultColor = Color.black;
    private Color color;
	private JButton colorButton = new JButton( "change color" );
	private ColorIcon colorIcon = new ColorIcon(32, 32, defaultColor);
	private JLabel swatch = new JLabel(colorIcon);
	private JLabel titleLabel = new JLabel();

	public final static String defaultPropertyName = "color";
	private String myPropertyName;

	private void setSwatchColor(Color c) {
		colorIcon.setColor(c);
		swatch.setIcon(colorIcon);
		swatch.repaint();
	}
	
    private boolean debug = true;
	private void dpl(String s) {
		if (debug) System.err.println(s);
	}
}
