/*
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package scotthaywood.ui;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.font.*;
import java.awt.geom.*;
import javax.swing.*;

/**
 * A <code>FontChooserBox</code> displays a list of all
 * the fonts available in the Java runtime environment,
 * with sample text rendered in each font.  It allows
 * a font to be selected by clicking on it, and allows
 * the sample text to be changed via a <code>JTextfield</code>.
 *
 * @author Scott Haywood
 * @version $Revision: 1.12 $
 */
public class FontChooserBox extends Box {
 	
    private boolean debug = true;
	private void dpl(String s) {
		if (debug) System.err.println(getClass().getName() + ": " +s);
	}

	/**
	 * Constructs a "font chooser box" with default settings:
	 * default title, no initially selected font, and a built-in
	 * value for the sample text.
	 * @see #FontChooserBox(String, Font, String)
	 * @see #setText
	 */
	public FontChooserBox() {
		this((Font) null, (String) null);
	}
	
	/**
	 * Constructs a font chooser box with a given font and
	 * sample text.
	 * @see #FontChooserBox(String, Font, String)
	 * @see #setText
	 */
	public FontChooserBox(Font f, String text) {
		this(null
		     /* "<html><center>Copyright 2002, Scott Haywood"+
		     "</center></html>"
		     */ ,
		     f, text);
	}
	/**
	 * Constructs a font chooser box.
	 * @param	title	string to be displayed at the top.
	 * @param	initialFont	font to be displayed as "selected"
	 * @param	text	sample text
	 *
	 */
	public FontChooserBox(String title, Font initialFont, String text) {
	
		super(BoxLayout.Y_AXIS);
		Box subBox = new Box(BoxLayout.X_AXIS);
		instanceCount++;
		getFamilyNames();
		if (title != null) {
			//dpl("adding title label");
			titleLabel = new JLabel(title);
			titleLabel.setFont(getInfoFont());
			javax.swing.border.Border border;
			border = BorderFactory.createCompoundBorder(
				BorderFactory.createLineBorder(Color.black),
				BorderFactory.createEmptyBorder(2,2,2,2)
				);
			JPanel tpanel = new JPanel();
			Box thbox = new Box(BoxLayout.X_AXIS);
			tpanel.setBorder(border);
			thbox.add(Box.createHorizontalGlue());
			thbox.add(titleLabel);
			thbox.add(Box.createHorizontalGlue());
			tpanel.add(thbox);
			add(tpanel);
		}
		if (initialFont != null) {
			int sz = initialFont.getSize();
			// note: negative font sizes not displayed correctly
			if (sz < minFontSize) {
				initialFont = initialFont.deriveFont(
					(float)(int)Math.max(minFontSize, Math.abs(sz)));
			}
			fontSize = initialFont.getSize();
			selectedFontIntValue = -1;
		}
		oldFont = selectedFont = initialFont;
		//dpl("initalFont="+initialFont+"\n"+
		//    "selectedFont=" + selectedFont);
		if (text != null)
			oldSampleText = sampleText = text;
		
		subBox.add(display);
		subBox.add(display.getScrollBar());
		add(subBox);
		sampleField = new JTextField(sampleText);
		sampleField.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				setText(sampleField.getText(), true);
				repaint();
			}
		});
		sampleField.setFont(getInfoFont());
		add(sampleField);
		
	}

	/**
	 * Frees the static list of families.
	 * This method is called when the object is
	 * about to be garbage-collected.
	 * @see java.lang.Object
	 */
	protected void finalize() throws Throwable {
		super.finalize();
		if (--instanceCount == 0 && freeMem)
			fams = null;
	}
	
	/**
	 * Gets the currently-selected font.
	 * Overrides Component.
	 *
	 * @returns the selected font.
	 */
	public Font getFont() { return selectedFont; }
	
	/**
	 * Sets the selected font and the sample text,
	 * without generating an event.
	 * @param	f	the new selected font
	 * @param	s	the new sample text
	 * @see #setEvent
	 * @see #setSelectedFont
	 * @see #setText
	 */
	synchronized public void setNoEvent(Font f, String s) {
		setSelectedFont(f, -1, false);
		sampleField.setText(s); // NOT expect to receive event
		setText(s, false);
	}
	
	/**
	 * Sets the selected font and the sample text,
	 * without generating an event.
	 * @param	f	the new selected font
	 * @param	s	the new sample text
	 * @see #setEvent
	 * @see #setSelectedFont
	 * @see #setText
	 */
	synchronized public void set(Font f, String s) {
		//dpl("set: s="+s);
		setSelectedFont(f, -1, true);
		sampleField.setText(s); // NOT expect to receive event
		setText(s, true);
		//dpl("set: sampleField text=" + sampleField.getText());
	}
	
	//
	// Sets the selected font, the numeric index of the
	// selected font, and optionally fires a property-change event.
	//
	private void setSelectedFont(Font f,
	                               int fontIntValue,
	                               boolean fireEvent) {
		oldFont = selectedFont;
		selectedFont = f;
		selectedFontIntValue = fontIntValue;
		//
		if (fireEvent &&
		    ( (f != null && !f.equals(oldFont)) ||
		      (f == null &&   oldFont != null)) )
		{
			firePropertyChange("font", oldFont, selectedFont);
			display.repaint();
		}
	}		

	/**
	 * Gets the sample text, a string which is rendered into
	 * each displayed font.
	 * @returns the sample text.
	 * @see #setText
	 * @see #getTypedText
	 */
	public String getText() {
		return sampleText;
	}
	
	/**
	 * Gets the text from the <code>JTextfield</code> which
	 * is used to change the sample text.
	 * This will differ from getText()
	 * if the user has not pressed the "Enter" key
	 * while typing in the <code>JTextfield</code>.
	 * @returns the text from the <code>JTextfield</code>.
	 * @see #getText
	 */
	public String getTypedText() {
		return sampleField.getText();
	}
	
	/**
	 * Sets the sample text and fires an event.
	 * @param	s	the new sample text.
	 * @see	#setText(String, boolean)
	 * @see #getText
	 */
	public void setText(String s) {
		setText(s, true);
	}

	/**
	 * Sets the sample text and optionally fires and event.
	 * @param	s	the new sample text
	 * @param	fireTextEvent	fires an event if true
	 * @see	#setText
	 * @see #getText
	 */
	protected void setText(String s, boolean fireTextEvent) {
		//dpl("setText: s=" + s);
		oldSampleText = sampleText;
		sampleText = s;
		//
		if (fireTextEvent &&
		    ( (s != null && !s.equals(oldSampleText)) ||
		      (s == null &&   oldSampleText != null)) )
		{
			//dpl("firing text change event");
			firePropertyChange("text", oldSampleText, sampleText);
		}
	}		
		
	
	// Rather than load the list of family names
	// when the class is loaded, it is loaded only when needed.
	// getFamilyNames() is called by each constructor that
	// needs the list of family names.
	// 
	static private String[] getFamilyNames() {
	
		if (fams == null) {
			GraphicsEnvironment ge =
        		GraphicsEnvironment.getLocalGraphicsEnvironment();
    		fams = ge.getAvailableFontFamilyNames();
		}
		return fams;
	}

	// Gets the font to be used for "informational purposes".
	// The font names are displayed in this font.
	// It should be a font which is legible.
	//
	private Font getInfoFont() {
		if (infoFont == null || infoFont.getSize() != infoFontSize)
			infoFont = new Font("Lucida Sans", Font.PLAIN, infoFontSize);
		return infoFont;
	}

    // ------------------------------------------

    // the number of objects of this class
    private static int instanceCount;
   	
   	private static boolean freeMem = true;
   	private static int minFontSize = 8;
	
   	// flag; true = display with antialiasing
   	private boolean antialias = true;
   	
    // the currently selected font, null if none.
    private Font selectedFont;

    // numeric index of the selected font, -1 if none
	private int selectedFontIntValue;
	
	// the previously selected font, used for generating
	// property-change events; null if none.
    private Font oldFont;

    // font size of the selected font and all displayed fonts
	private int fontSize = 32;
	
	// size of the "informational" font
	private int infoFontSize = 16;
	
	// the "informational" font.
	private Font infoFont;
	
	// the sample text.  It is rendered into each displayed font.
	private String sampleText = "Sample Text 123";
	
	// old value of the sample text, for generating events.
	private String oldSampleText = sampleText;
	
	// text field for changing the sample text
	private JTextField sampleField;
	
	// label displayed at the top of the box
    private JLabel titleLabel = new JLabel();

    // the list of currently displayed fonts
	private ArrayList displayedFonts = new ArrayList();
	
	// the list of font families in the Java environment
    private static String[] fams;

    // font styles
	protected final int[] styles = { Font.PLAIN, Font.BOLD, Font.ITALIC,
						   Font.BOLD + Font.ITALIC };
	
	// a cache of fonts.
	// There may be too many fonts
	// in the system to keep them all in memory at one time.					   					
	private FontCache fontCache = new FontCache();
	
	// A type of JPanel for displaying a subset of the
	// fonts.  See the class definition.
	private DisplayPanel display = new DisplayPanel();
	
	// A convenience structure that for associating a font
	// with its position within the display, and with
	// an index into a hypothetical array of all fonts.
	private class DisplayedFont {
		public Rectangle2D rect;
		public Font font;
		public int fontIntValue;
		public DisplayedFont(FontPointer fp, Rectangle2D r) {
			rect = r;
			fontIntValue = fp.intValue();
			font = fp.getFont();
		}
	}
	
	// The FontCache supplies fonts in a way that balances
	// two goals: minimizing memory usage, and minimizing
	// the re-construction of Font objects.
	//
	private class FontCache {

		public FontCache() { 
			getFamilyNames();
			//map = new HashMap((int)Math.round(maxCacheFonts/0.75) + 1);
			map = new HashMap((int)Math.round(fams.length/0.75) + 1);
		}
		
		// gets an array of all the fonts in a family
		//
		public Font[] getFamily(int ndx, int size) {
			//Entry e = new Entry(ndx, size, null);
			tmpEntry.famIndex = ndx;
			tmpEntry.size = size;
			Entry existing = (Entry) map.get(tmpEntry);
			if (existing != null) {
				//dpl("1 - font family already in cache");
				return existing.fonts;
			}
			else {
				//dpl("2 - adding font family to cache");
				Entry e = new Entry(ndx, size, null);
				return add(e).fonts;
			}
		}

		// gets a font of a given family, style, and size
		//
		public Font getFont(int ndx, int styleIndex, int size) {
			Font[] fontarray = getFamily(ndx, size);
			return fontarray[styleIndex];
		}

		
		// An entry in the font cache,
		// which contains all the valid styles
		// for a specific family.
		// see add(Entry)
		//
		private class Entry {
			int famIndex;
			int size;
			Font[] fonts;
			public Entry(int ndx, int sz, Font[] fa) {
				famIndex = ndx;
				size = sz;
				fonts = fa;
			}
			public int hashCode() {
				return famIndex;
			}
			public boolean equals(Object obj) {
				if (obj == this)
					return true;
				if (obj != null) {
					try {
						Entry e = (Entry)obj;
						return e.famIndex == famIndex &&
						       e.size == size;
					}
					catch (ClassCastException ex) {}
				}
				return false;
			}
		}

		// Create of the given family, style, and size.
		// Returns null if that exact font cannot be constructed.
		//
		private Font createFont(String fam, int style, int size) {
			Font f = new Font(fam, style, size);
			return (f.getName().equals(fam) && 
			        f.getStyle() == style &&
			        f.getSize() == size)
				   ? f : null;
		}

		// Adds an entry to the cache.
		// Uses the family number and size of the given object,
		// and fills all the valid fonts (of different styles),
		//
		private Entry add(Entry e) {
			int ndx = e.famIndex;
			int size = e.size;
			Font[] fa = new Font[styles.length];
			Font f;
			for (int st = 0; st<styles.length; st++) {
				fa[st] = createFont(fams[ndx], styles[st], size);
			}

			// get rid of duplicate fonts
			for (int i=0; i<fa.length; i++) {
				if (fa[i] == null) continue;
				for (int j=i+1; j<fa.length; j++) {
					if (fa[j] == null) continue;
					if ( fa[i].getFontName().equals(fa[j].getFontName()) )
						fa[j] = null;
				}
			}

			if (recent.size() + fa.length > maxCacheFonts) {
				Entry oldEntry = (Entry)recent.getLast();
				map.remove(oldEntry);
				recent.removeLast();
			}
			e.fonts = fa;
			recent.addFirst(e);
			map.put(e, e);
			return e;
		}

		private LinkedList recent = new LinkedList();
		private HashMap map;
		private int maxCacheFonts = 40;
		//private int maxCacheFonts = 800;
		private Entry tmpEntry = new Entry(0,0,null);
		
	}

	/*
	*	A FontPointer represents a pointer into an imaginary
	*	array of Fonts, arranged by font family, and within
	*	each family, by font style.  Some
	*	family/style combinations may not exist.
	*
	*	The public methods for changing the pointer, such as
	*	incr() and decr(), will skip null values.
	*/
	private class FontPointer {
		private int fam;
		private int style;
		private Font font;
		private transient FontPointer tmp;

		// point to the first font
		public FontPointer() {
			//dpl("Enter FontPointer, fam="+ fam + ",style=" + style); 
			getFamilyNames();
			font = fontCache.getFont(fam, style, fontSize); 
			if (font == null)
				incr();
			tmp = new FontPointer(false);
			//dpl("FontPointer(): " + this);
		}
		public FontPointer(FontPointer fp) {
			setPointer(fp);
			tmp = new FontPointer(false);
		}
		private FontPointer(boolean flag) {}
		
		private void setPointer(FontPointer fp) {
			fam = fp.fam;
			style = fp.style;
			font = fp.font;
		}

		public String toString() {
			return "[fam="+fam+",style="+style+"]";
		}

		// return a numeric index
		// which represents the font to which this points.
		//
		public int intValue() {
			return fam * styles.length + style;
		}
		
		public boolean equals(Object obj) {
			//dpl("Enter equals");
			FontPointer fp = (FontPointer)obj;
			if (this == fp)
				return true;
			return fam == fp.fam && style == fp.style;
		}
		
		public int getFamIndex() { return fam; }
		public int getStyleIndex() { return style; }

		public Font getFont() {
			return font;
		}

		/*
		*	Return an iterator for the font list.
		*	The next() method of the iterator returns
		*	a pointer into the list; the caller should
		*	not modify this pointer (don't call its incr()
		*	method, for example).
		*/
		public Iterator iterator() {
			return new Iterator() {
				FontPointer fp = new FontPointer(FontPointer.this);
				boolean pastEnd = false;

				public boolean hasNext() {
					if (pastEnd) return false;
					return fp.font != null;
				}
				public Object next() {
					Font f = fp.font;
					if (pastEnd || f==null) 
						throw new NoSuchElementException();
					if (!fp.incr())
						pastEnd = true;
					return fp;
				}
				public void remove() 
					throws UnsupportedOperationException
				{
					throw new UnsupportedOperationException();
				}
			};
		}

		// true if pointing to the last slot in the array
		private boolean atEnd() {
			return (fam == fams.length - 1) && 
				   (style == styles.length - 1);
		}

		// point to the last slot in the array
		private void setToEnd() {
			fam = fams.length - 1;
			style = styles.length - 1;
		}

		// true if pointing to the first slot in the array
		private boolean atBeginning() {
			return fam == 0 && style == 0;
		}

		// point to the first slot in the array
		private void setToBeginning() {
			fam = 0;
			style = 0;
		}

		/* If there is a next font, point to it, and return true.
		*  Otherwise, return false.
		*  This uses fontCache, fontSize. Not thread-safe.
		*/
		public boolean incr() {
			tmp.setPointer(this);
			while (!atEnd()) {
				if (++style >= styles.length) {
					style = 0;
					if (++fam >= fams.length) {
						setToEnd();
					}
				}
				//dpl("incr: this=" + toString());
				font = fontCache.getFont(fam, style, fontSize); 
				if (font != null) {
					return true;
				}
			}
			setPointer(tmp);
			return false;
		}

		/* If there is a previous font, point to it, and return true.
		*  Otherwise, return false.
		*  This uses fontCache, fontSize. Not thread-safe.
		*/
		public boolean decr() {
			tmp.setPointer(this);
			while (!atBeginning()) {
				if (--style < 0) {
					style = styles.length - 1;
					if (--fam < 0) {
						setToBeginning();
					}
				}
				font = fontCache.getFont(fam, style, fontSize);
				if (font != null) {
					return true;
				}
			}
			setPointer(tmp);
			return false;
		}

		// Move ahead in the font list by 'n' fonts.
		// If there are fewer than n non-null fonts following
		// the current one, this will point to the last non-null font.
		//
		public void incr(int n) {
			for (; n>0 ;n--) {
				if (!incr())
					break;
			}
		}

		// Move back in the font list by 'n' fonts.
		// If there are fewer than n non-null fonts preceding
		// the current one, this will point to the first non-null font.
		//
		public void decr(int n) {
			for (; n>0 ;n--) {
				if (!decr())
					break;
			}
		}

		public void change(int n) {
			if (n < 0)
				decr(Math.abs(n));
			else
				incr(n);
		}
	}
	
	// A type of JPanel for displaying fonts
	//
	private class DisplayPanel extends JPanel {
		public DisplayPanel() {

			getFamilyNames();
			int sb_extent = 1;
			int sb_max = fams.length * styles.length;
			sbar = new JScrollBar(SwingConstants.VERTICAL,
							   0, sb_extent, 0, sb_max);
			sbar_old_value = sbar.getValue();
			sbar.addAdjustmentListener(new AdjustmentListener() {
				public void adjustmentValueChanged(AdjustmentEvent e) {
					int val = sbar.getValue();
					//dpl("Enter adjust: val=" + val);
					if (sbar_old_value == val) {
						//dpl("adjust: no change");
						return;
					}
					FontPointer old = new FontPointer(firstFontOfDisplay);
					//dpl("adjustment type = " + e.getAdjustmentType()); 
					//dpl("sbar block = " + sbar.getBlockIncrement()); 
					//dpl("sbar value = " + sbar.getValue()); 
					//dpl("e value = " + e.getValue()); 

					// note: it seems we always get TRACK adjustments,
					//    regardless of the type of change.
					switch (e.getAdjustmentType()) {
					case AdjustmentEvent.UNIT_INCREMENT :
						firstFontOfDisplay.incr();
						break;
					case AdjustmentEvent.UNIT_DECREMENT :
						firstFontOfDisplay.decr();
						break;
					case AdjustmentEvent.BLOCK_INCREMENT  :
						firstFontOfDisplay.incr(displayedFonts.size());
						break;
					case AdjustmentEvent.BLOCK_DECREMENT :
						firstFontOfDisplay.decr(displayedFonts.size());
						break;
					case AdjustmentEvent.TRACK :
						int diff = val - sbar_old_value;
						//dpl("old pointer: " + firstFontOfDisplay);
						firstFontOfDisplay.change(diff);	
						//dpl("new pointer: " + firstFontOfDisplay);
						break;
					default :
						//dpl("adjustment type = " + e.getAdjustmentType());
						return;
					}
					sbar_old_value = val;
					//dpl("adjust: doing repaint()");
					if (!firstFontOfDisplay.equals(old))
						repaint();
				}
			});

			addMouseListener(new MouseAdapter() {
				public void mousePressed(MouseEvent e) {
					mouse1 = e.getPoint();
				}
				public void mouseReleased(MouseEvent e) {
					mouse2 = e.getPoint();
					checkMousePoints();
				}
				public void mouseEntered(MouseEvent e) {
					mouse1 = mouse2 = null;
				}

			});
			
			setPreferredSize(new Dimension(prefWidth, prefHeight));
			setBorder(BorderFactory.createEmptyBorder(4,4,4,4));
		}

		Dimension screenDim = Toolkit.getDefaultToolkit().getScreenSize();
		int prefWidth = (int) Math.round(screenDim.width * 0.25);
		int prefHeight = (int) Math.round(screenDim.height * 0.75);

		private Point mouse1, mouse2;
		JScrollBar sbar;
		int sbar_old_value;
		FontPointer firstFontOfDisplay = new FontPointer();

		private void checkMousePoints() {
			if (mouse1 == null || mouse2 == null)
				return;
			for (int i = 0; i < displayedFonts.size(); i++) {
				DisplayedFont df = (DisplayedFont) displayedFonts.get(i);
				Rectangle2D r = df.rect;
				if (r.contains(mouse1) && r.contains(mouse2)) {
					setSelectedFont(df.font, df.fontIntValue, true);
					break;
				}
			}
		}

        // Get the scrollbar for scrolling thru the list of fonts.
        // Note that the DisplayPanel does not display the scrollbar.
        //
		public JScrollBar getScrollBar() {
			return sbar;
		}

		// Reverse the colors of a graphics object.
		private void reverseColors(Graphics2D g) {
			Color c = g.getColor();
			g.setColor(g.getBackground());
			g.setBackground(c);
		}

		// Paint this panel.  Overrides JComponent.
		public void paintComponent(Graphics graphics) {

			super.paintComponent(graphics);
			Graphics2D g = (Graphics2D) graphics;
			
			g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                             (antialias
                              ? RenderingHints.VALUE_ANTIALIAS_ON
                              : RenderingHints.VALUE_ANTIALIAS_OFF));
        	g.setRenderingHint(RenderingHints.KEY_RENDERING,
                             (antialias
                              ? RenderingHints.VALUE_RENDER_QUALITY
                              : RenderingHints.VALUE_RENDER_SPEED));
			
			
			g.setBackground(getBackground());
			g.clearRect(0,0,getWidth(), getHeight());
			Insets insets = getBorder().getBorderInsets(this);
			int w = getWidth() - insets.left - insets.right ;
			int h = getHeight() - insets.top - insets.bottom;
			g.translate(insets.left, insets.top);
			g.setClip(0, 0, w, h);
			//g.setClip(insets.left, insets.top, w, h);

			Font ifont = getInfoFont();
			FontRenderContext frc = g.getFontRenderContext();
			displayedFonts.clear();

			Line2D.Float horiz = new Line2D.Float();
			//dpl("paintComponent: firstFontOfDisplay = "+firstFontOfDisplay);
			Iterator it = firstFontOfDisplay.iterator();
			for (int vert=4; it.hasNext() && vert<h ; ) {

				FontPointer fp = (FontPointer) it.next();
				Font f = fp.getFont();
				
				LineMetrics ilm, lm;
				ilm = ifont.getLineMetrics(f.getFontName(), frc);
				lm = f.getLineMetrics(sampleText, frc);

				// remember the display location of this font.
				Rectangle2D.Float rect = new Rectangle2D.Float(
					insets.left, insets.top + vert,
					w, ilm.getHeight() + lm.getHeight());

				if (vert + rect.height > h)
					break;

				displayedFonts.add(new DisplayedFont(fp, rect));

				boolean reverse = false;
				if (fp.intValue() == selectedFontIntValue ||
				    (selectedFontIntValue == -1 && 
					 fp.getFont().equals(selectedFont)) )
				{
					reverse = true;
					Rectangle2D.Float r2 = new Rectangle2D.Float(
						0, vert, rect.width, rect.height);
					g.fill(r2);
					reverseColors(g);
				}

				// display the name of the font, and sample text.
				// The "leading" (LineMetrics.getLeading()) is put
				// below each line.
				g.setFont(ifont);
				g.drawString(f.getFontName(), 
				             0, 
							 vert + ilm.getAscent());
				g.setFont(f);
				g.drawString(sampleText,
							 0, 
							 vert + rect.height 
							  - lm.getLeading() - lm.getDescent());

				if (reverse)
					reverseColors(g);

				horiz.setLine(0,vert+rect.height, w,vert+rect.height);
				g.draw(horiz);
				vert += Math.round(rect.height + 1);
			}
			//dpl("displayedFonts.size()="+displayedFonts.size());

			if (sbar.getBlockIncrement() != displayedFonts.size()) {
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						int newBlockIncr = displayedFonts.size();
						if (newBlockIncr == 0) newBlockIncr = 1;
						//dpl("setting block increment to " + newBlockIncr); 
						sbar.setBlockIncrement(newBlockIncr);
					}
				});
			}
			int pval = firstFontOfDisplay.intValue();
			int val = sbar.getValue();
			//dpl("pval=" + pval + ", val=" + val);
			if (val != pval) {
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						sbar_old_value = firstFontOfDisplay.intValue();
						//dpl("setting sbar to " + sbar_old_value);
						sbar.setValue(sbar_old_value);
					}
				});
			}
		}
	}
}
