/*
Copyright (C) 2002, Scott Haywood

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package scotthaywood.ui;

import java.util.*;
import java.awt.*;
import java.awt.font.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;

/**
 * A <code>PointSelector</code>
 * displays a component and allows a point within
 * the component to be selected interactively.
 * It generates property-change events when the point changes.
 *
 * @author Scott Haywood
 * @version $Revision: 1.8 $
 */
public class PointSelector extends JPanel {

	private void dpl(String s) {
		System.err.println(getClass().getName() + ": " + s);
	}

	private int round(double d) {
   		return (int)Math.round(d);
   	}

	static double inch(double x) {
		return x * PIXELS_PER_INCH;
	}

	//
	final static int PIXELS_PER_INCH =
		Toolkit.getDefaultToolkit().getScreenResolution();
	
	private Component target;
	private JLayeredPane layers = new JLayeredPane();
	private ToolPanel tools = new ToolPanel();
	//private String tip = "Blah blah blah blah ...";
	private String tip = "move the sliders to select a different point";

	private Point selected = new Point();
	private Point oldSelected = new Point();
	private float xRatio;
	private float yRatio;

	private int trackWidth = round(inch(0.25));
	private int thumbWidth = trackWidth;
	private Rectangle frameRect = new Rectangle();
	private Rectangle viewRect = new Rectangle();
	private Area frameBorder;
	private Rectangle vertTrack = new Rectangle();
	private Rectangle horizTrack = new Rectangle();
	private Rectangle vertThumb = new Rectangle();
	private Rectangle selectedThumb = null;
	private Rectangle horizThumb = new Rectangle();
	private Rectangle cornerRect = new Rectangle();
	private int defaultViewSize = 100;
	private boolean resizeTarget = true;
	//private Color frameColor1 = new Color(96,96,96);
	private Color frameColor1 = new Color(128,128,128);
	private Color frameColor2 = frameColor1.darker();
	private Color trackColor = new Color(24,120,24);
	private Color cornerColor = trackColor;
	private Color thumbEnabledColor = new Color(128,128,128);
	private Color thumbDisabledColor = trackColor;
	private Color lineColor1 = Color.black;
	private Color lineColor2 = Color.white;
	private boolean enabled = true;
	private boolean crossHairsEnabled = true;
	private boolean reportingPointOnResize = false;

	/**
	 * Changes whether the point can be selected.
	 * @param 	b	if true, selecting the point is enabled,
	 * otherwise it is disabled.
	 * @see #isEnabled
	 */
	public void setEnabled(boolean b) {
		boolean old = enabled;
		enabled = b;
		setToolTipText(enabled ? tip : null);
		if (old != enabled)
			repaint();
	}
	
	/**
	 * Reports whether the point can be selected.
	 * @returns true if the point can be selected, false if not.
	 * @see #setEnabled
	 */
	public boolean isEnabled() {
		return enabled;
	}
	
	/**
	 * Sets whether a property-change event is generated
	 * when the component is resized.
	 * @param	b	if true, property-change events will be
	 * generated when the component is resized.
	 * @see #isReportingPointOnResize
	 */
	public void setReportingPointOnResize(boolean b) {
		reportingPointOnResize = b;
	}
	
	/**
	 * Tells whether a property-change event is generated
	 * when the component is resized.
	 * @returns true if property-change events will be
	 * generated when the component is resized.
	 * @see #setReportingPointOnResize
	 */
	public boolean isReportingPointOnResize() {
		return reportingPointOnResize;
	}

	// Calculate the ratios of the selected point's coordinates
	// to the dimensions of the view rectangle.
	private void setRatios() {
		if (target == null) {
			xRatio = (float) selected.x / viewRect.width;
			yRatio = (float) selected.y / viewRect.height;
		}
		else {
			xRatio = (float) selected.x / target.getWidth();
			yRatio = (float) selected.y / target.getHeight();
		}
	}

	// return the view width based on the width of this object
	private int calcViewWidth() {
		return calcViewWidth(getWidth());
	}
	// return the view height based on the height of this object
	private int calcViewHeight() {
		return calcViewHeight(getHeight());
	}
	// return the view width to be used, given a width
	// in which this object might be displayed.
	private int calcViewWidth(int total_w) {
		int w = total_w - 3 * trackWidth;
		return w < 10 ? 10 : w;
	}
	// return the view height to be used, given a height
	// in which this object might be displayed.
	private int calcViewHeight(int total_h) {
		int h = total_h - 3 * trackWidth;
		return h < 10 ? 10 : h;
	}
	// return the width needed to display a
	// component in a viewing area of a given width
	private int calcWidth(int view_width) {
		return view_width + 3 * trackWidth;
	}
	// return the height needed to display a
	// component in a viewing area of a given height
	private int calcHeight(int view_height) {
		return view_height + 3 * trackWidth;
	}

	// set the locations and sizes of various parts of the display
	private void setDimensions() {
		int	w = calcViewWidth();
		int	h = calcViewHeight();
		//dpl("setDim: w="+ w +", h="+ h);
		frameRect.setRect(0,0, w + 2 * trackWidth,
		                       h + 2 * trackWidth);
		viewRect.setRect(frameRect.x + trackWidth,
		                 frameRect.y + trackWidth,
						 w,h);
		vertTrack.setRect(frameRect.width,
		                  0,
		                  trackWidth,
						  frameRect.height);
		frameBorder = new Area(frameRect);
		frameBorder.subtract(new Area(viewRect));
		horizTrack.setRect(0,
		                   frameRect.height,
		                   frameRect.width,
						   trackWidth);
		cornerRect.setRect(frameRect.width,
		                  frameRect.height, 
						  trackWidth, 
						  trackWidth);
	}

	// set the locations and sizes of the "thumbs" (the images
	// which can be slid along tracks to move the selected point).
	private void setThumbs() {
		setVertThumb();
		setHorizThumb();
	}

	// set the location and size of the vertical thumb.
	private void setVertThumb() {
		vertThumb.setRect(
			vertTrack.x, viewRect.y + selected.y - thumbWidth/2,
			trackWidth, thumbWidth
			);
		//dpl("vertThumb = " + vertThumb);
	}

	// set the location and size of the horizontal thumb.
	private void setHorizThumb() {
		horizThumb.setRect(
			viewRect.x + selected.x - thumbWidth/2, horizTrack.y,
			thumbWidth, trackWidth
			);
	}

	
	/**
	 * Get the target component
	 * @returns the target component
	 */
	public Component getTarget() {
		return target;
	}

	/**
	 * Gets the selected point.
	 * @returns the selected point.
	 * @see #setPoint
	 */
	public Point getPoint() {
		return new Point(selected);
	}

	/**
	 * Sets the selected point.  Generates a property-change event.
	 * @param	p	 the new selected point.
	 * @see #getPoint
	 * @see #setPoint(int, int)
	 */
	public void setPoint(Point p) {
		setPoint(p.x, p.y);
	}

	/**
	 * Sets the selected point.  Generates a property-change event.
	 * @param	x	 the new horizontal coordinate.
	 * @param	y	 the new vertical coordinate.
	 * @see #getPoint
	 * @see #setPoint(Point)
	 */
	public void setPoint(int x, int y) {
		setPoint0(x,y,true);
		setRatios();
	}
	
	// Set the selected point, optionally generating
	// a property-change event.
	private void setPoint0(int x, int y, boolean fireEvent) {

		Component c = target;
		if (c != null) {
			if (x >= c.getWidth()) x = c.getWidth() - 1;
			if (y >= c.getHeight()) y = c.getHeight() - 1;
		}
		if (x < 0) x = 0;
		if (y < 0) y = 0;

		if (selected.x != x || selected.y != y) {
			oldSelected.setLocation(selected);
			selected.x = x;
			selected.y = y;
			//dpl("new x="+x+", new y="+y);

			// to do: copy the values before passing to listener
			//
			if (fireEvent)
				firePropertyChange("point", oldSelected, selected);

			repaint();
		}
	}

	/*
	public PointSelector() {
		this(null, 0, 0);
	}
	*/

	/**
	 * Constructs a point selector with a given target.
	 * The selected point is initially set the the middle
	 * of the target.
	 * @param	c	the target component
	 */
	public PointSelector(Component c) {
		this( c, c.getWidth()/2, c.getHeight()/2 );
	}

	/**
	 * Constructs a point selector
	 * @param	c	the target component
	 * @param	initialSelection	the initially-selected point
	 */
	public PointSelector(Component c, Point initialSelection) {
		this(c, initialSelection.x, initialSelection.y);
	}

	/**
	 * Constructs a point selector
	 * @param	c	the target component
	 * @param	x	the initially-selected horizontal coordinate
	 * @param	y	the initially-selected vertical coordinate
	 */
	public PointSelector(Component c, int x, int y) {

		// normalize and save the parameters
		//
		if (c != null) {
			if (x >= c.getWidth()) x = c.getWidth() - 1;
			if (y >= c.getHeight()) y = c.getHeight() - 1;
		}
		if (x < 0 ) x = 0;
		if (y < 0 ) y = 0;
		oldSelected.x = selected.x = x;
		oldSelected.y = selected.y = y;

		target = c;

		// Set the size
		//
		int w, h;
		if (target == null) {
			w = h = defaultViewSize;
		}
		else {
			Dimension d = target.getSize();
			w = d.width;
			h = d.height;
		}
		Dimension d = new Dimension( calcWidth(w), calcHeight(h) );
		setSize(d);
		setPreferredSize(d);
		//dpl("constructor: size="+getSize());
		tools.setSize(d);
		layers.setSize(d);
		
		// Set the location and size of the various rectangles
		// which describe the layout, and which are used by
		// ToolPanel.paintComponent().
		// One of them is viewRect, which is also used to
		// place the target.
		//
		setDimensions();
		
		// Add components.
		// This object will have a JLayeredPane,
		// which contains the target (at layer 0)
		// and the ToolPanel (at layer 1).
		//
		setLayout(null);
		layers.setLocation(0,0);
		add(layers);
		if (target != null) {
			target.setLocation(viewRect.x, viewRect.y);
			layers.add(target, new Integer(0));
		}
		tools.setLocation(0,0);
		layers.add(tools, new Integer(1));

		
		// Remember the relative position of the selected point.
		// Note this depends on the sizes of viewRect and target,
		// so it must be called after setDimensions.
		//
		setRatios();

		// Make sure the tool tip, etc. is consistent
		// with the value of enabled.
		//
		setEnabled(enabled);

		// Set the event listeners
		//
		addMouseListener(mia);
		addMouseMotionListener(mia);
		addComponentListener(new ComponentAdapter() {
			public void componentResized(ComponentEvent e) {
				//dpl("----------");
				
				//setPreferredSize(getSize());

				int vw = calcViewWidth();
				int vh = calcViewHeight();
				int x,y;
				if (target != null) {
				
					// Resize the target
					//
					int newTargetWidth = vw;
					int newTargetHeight = vh;
					if (resizeTarget) {
						target.setSize(newTargetWidth, newTargetHeight);
						//dpl("resized target, size="+target.getSize());
					}

					// New selected point is relative to the target size
					//
					x = round(xRatio * newTargetWidth);
					y = round(yRatio * newTargetHeight);
				}
				else {
					// New selected point is relative to the view size
					//
					x = round(xRatio * vw);
					y = round(yRatio * vh);
				}
				
				// Resize the JLayeredPane and ToolPanel
				// to match this object.
				//
				layers.setSize(getSize());
				tools.setSize(getSize());
				
				// New selected point is always within the view
				//
				if (x >= vw) x = vw - 1;
				if (y >= vh) y = vh - 1;

				setDimensions();
				//dpl("new view w="+viewRect.width+", old x="+selected.x);
				//dpl("new view h="+viewRect.height+", old y="+selected.y);
				setPoint0(x, y, reportingPointOnResize);
				if (false) dpl("--- x="+x+", y="+y+
				    ", xRatio="+xRatio+
					", yRatio="+yRatio+" ---");
			}
		});
	}

	private MouseInputAdapter mia = new MouseInputAdapter() {

		Point pressed = null;
		Point psel = new Point(); // the selected point when mouse pressed
		Cursor oldCursor = null;

		public void mouseClicked(MouseEvent e) {
			if (!enabled) return;
			Point p = e.getPoint();
			Rectangle[] r = new Rectangle[] {
				vertTrack, vertThumb, horizTrack, horizThumb
			};
			boolean changed = false;
			int x = 0, y = 0;
			for (int i=0; i<r.length; i+=2) {
				if (r[i].contains(p)) {
					x = selected.x;
					y = selected.y;
					Rectangle th = r[i+1];
					if (r[i] == vertTrack) {
						if (p.y < th.y) {
							y--; changed = true;
						}
						else if (p.y >= th.y + th.height) {
							y++; changed = true;
						}
					}
					else {
						if (p.x < th.x) {
							x--; changed = true;
						}
						else if (p.x >= th.x + th.width) {
							x++; changed = true;
						}
					}
					break;
				}
			}
			if (changed) setPoint(x,y);
		}
		public void mousePressed(MouseEvent e) {
			if (!enabled) return;
			Point p = e.getPoint();
			Rectangle[] r = new Rectangle[] {vertThumb, horizThumb};
			for (int i=0; i<r.length; i++) {
				if (r[i].contains(p)) {
					pressed = p;
					psel.setLocation(selected);
					selectedThumb = r[i];
					break;
				}
			}
		}
		public void mouseDragged(MouseEvent e) {
			if (!enabled) return;
			if (pressed == null) return; // should not happen
			Point p = pressed;
			Point q = e.getPoint();
			//dpl("p="+p+", q="+q);

			int x = psel.x;
			int y = psel.y;
			int w = viewRect.width;
			int h = viewRect.height;
			if (target != null) {
				/* temporary tests
				if (w != target.getWidth())
					throw new Error("w ??");
				if (h != target.getHeight())
					throw new Error("h ??");
				*/
				
				// The viewRect size should be the same as the target. size.
				// This code is needed only if
				// resizeTarget is false (which has not been tested
				// as of 12 Nov 2001).	
				w = Math.min(w, target.getWidth());
				h = Math.min(h, target.getHeight());
			}
			if (selectedThumb == horizThumb) {
				x += q.x - p.x;
				if (x < 0)
					x = 0;
				else if (x > w)
					x = w - 1;
			}
			else if (selectedThumb == vertThumb) {
				y += q.y - p.y;
				if (y < 0)
					y = 0;
				else if (y > h)
					y = h - 1;
			}
			setPoint(x,y);
		}
		public void mouseReleased(MouseEvent e) {
			pressed = null;
			selectedThumb = null;
			if (!enabled) return;
			//PointSelector.this.setCursor(oldCursor);
		}
		
	};

	
	// A ToolPanel displays the frame around the target component,
	// and the various parts for selecting the point.
	// It is layered above the target component.
	private class ToolPanel extends JPanel {
		{
			setOpaque(false);
		}
		public void paintComponent(Graphics g) {
			paintToolPanel(g);
		}
	}
	
	// paint all of the parts except the target component
	//
	private void paintToolPanel(Graphics graphics) {

		int w = getWidth();
		int h = getHeight();
		int min = Math.min(w,h);
		//float thickness = crossThickness;
		//if (thickness > min/2f) thickness = min/2f;

		Graphics2D g = (Graphics2D) graphics;

		Color oldColor = g.getColor();

		// draw the corner
		//
		//dpl("cornerRect="+cornerRect);
		g.setColor(cornerColor);
		g.fill(cornerRect);

		// draw the tracks
		//
		g.setColor(trackColor);
		g.fill(horizTrack);
		g.fill(vertTrack);

		// draw the frame
		//
		drawFrame(g);

		// draw the thumbs
		//
		if (enabled)
			drawThumbs(g);

		// draw the crosshairs
		//
		if (enabled && crossHairsEnabled) {
			g.setColor(oldColor);
			drawLines(g);
		}

		// restore the Graphics object
		//
		g.setColor(oldColor);

	}

	//
	// Draw the cross-hairs
	//
	private void drawLines(Graphics2D g) {

		Shape oldClip = g.getClip();
		g.clip(viewRect);

		if (target != null) {
			g.translate(viewRect.x, viewRect.y);


			int w = target.getWidth();
			int h = target.getHeight();

			// get the target's image
			GraphicsConfiguration gconf = g.getDeviceConfiguration();
			BufferedImage bi;
			bi = gconf.createCompatibleImage(w, h);
			//dpl("bi="+bi+", target size = "+target.getSize());
			Graphics2D g2 = bi.createGraphics();
			GeneralPath gp = new GeneralPath(new Rectangle(
				0, selected.y, w, 1));
			gp.append(new GeneralPath(new Rectangle(
				selected.x, 0, 1, h)), false);
			g.clip(gp);
			target.paint(g2);

			// add the lines
			//
			// horizontal line
			int[] b = new int[w];
			bi.getRGB(0, selected.y, w, 1, b, 0, w);
			invertColors(b);
			bi.setRGB(0, selected.y, w, 1, b, 0, w);
			//
			// vertical line
			b = new int[h];
			bi.getRGB(selected.x, 0, 1, h, b, 0, 1);
			invertColors(b);
			bi.setRGB(selected.x, 0, 1, h, b, 0, 1);

			// draw the target
			g.drawImage(bi, null, 0, 0);

			g.translate(-viewRect.x, -viewRect.y);
		}
		else { // no target
			// .......................
		}
		g.setClip(oldClip);
	}

	// invert the colors as represented in an array of integers
	private void invertColors(int[] buf) {
		long x;
		//dpl("---");
		for (int i=0; i<buf.length; i++) {
			int a,b,c,d;
			a = (buf[i] >> 24) & 0xff;
			b = (buf[i] >> 16) & 0xff;
			c = (buf[i] >> 8 ) & 0xff;
			d = (buf[i]      ) & 0xff;
			/*
			dpl("rgb ["+i+"] = " + Integer.toHexString(buf[i]) +
			    " = " + a + 
			    " " + b + 
			    " " + c + 
			    " " + d );
			*/
			x = buf[i] & 0xff000000;
			x += ((255 - b) << 16);
			x += ((255 - c) <<  8);
			x += ((255 - d)      );
			buf[i] = (int) x;
			//dpl(" new val = " + Integer.toHexString(buf[i]));
		}
		//dpl("---");
	}

	// draw the frame around the target component
	//
	private void drawFrame(Graphics2D g) {

		Shape oldClip = g.getClip();
		g.setClip(frameBorder);

		Rectangle2D.Float r = new Rectangle2D.Float();
		g.setColor(frameColor1);
		g.fill(frameRect);
		g.setColor(frameColor2);

		int nn;
		float len;

		// draw the vertical bands across the width of the view
		//
		nn = viewRect.width / trackWidth;
		if ((nn & 1) == 0) {
			// decrement to odd number, but at least 1
			nn = (nn==0) ? 1 : nn-1;
		}
		len = (float)viewRect.width / nn;
		for (int i=0; i<nn; i+=2) {
			r.setRect(
				viewRect.x + i * len, frameRect.y,
				len, frameRect.height
				);
			g.fill(r);
		}

		// draw the horizontal bands across the height of the view
		//
		nn = viewRect.height / trackWidth;
		if ((nn & 1) == 0) {
			// decrement to odd number, but at least 1
			nn = (nn==0) ? 1 : nn-1;
		}
		len = (float)viewRect.height / nn;
		for (int i=0; i<nn; i+=2) {
			r.setRect(
				frameRect.x, viewRect.y + i * len,
				frameRect.width, len
				);
			g.fill(r);
		}

		// restore
		//
		g.setClip(oldClip);
	}

	
	// return a triangle which is the upper left half
	// of the given rectangle
	private Polygon upper_left_trangle(Rectangle r) {

		Polygon t = new Polygon();
		t.addPoint(r.x, r.y);
		t.addPoint(r.x + r.width - 1, r.y);
		t.addPoint(r.x, r.y + r.height - 1);
		return t;
	}
	
	// return a triangle which is the lower right half
	// of the given rectangle
	private Polygon lower_right_trangle(Rectangle r) {

		Polygon t = new Polygon();
		t.addPoint(r.x + r.width - 1, r.y);
		t.addPoint(r.x + r.width - 1, r.y + r.height - 1);
		t.addPoint(r.x, r.y + r.height - 1);
		return t;
	}

	// paint the "thumbs", with are the images that can
	// be slid on tracks to select the point.
	//
	private void drawThumbs(Graphics2D g) {
		setThumbs();
		Color c = enabled ? thumbEnabledColor : thumbDisabledColor;

		Polygon triangle;
		g.setColor(c.brighter());
		triangle = upper_left_trangle(vertThumb);
		g.fill(triangle);
		triangle = upper_left_trangle(horizThumb);
		g.fill(triangle);

		g.setColor(c.darker());
		triangle = lower_right_trangle(vertThumb);
		g.fill(triangle);
		triangle = lower_right_trangle(horizThumb);
		g.fill(triangle);

		g.setColor(c);
		Rectangle r = vertThumb;
		g.fillRect(r.x + 2, r.y + 2, r.width - 4, r.height - 4);
		r = horizThumb;
		g.fillRect(r.x + 2, r.y + 2, r.width - 4, r.height - 4);

	}

}
